/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "gl_state.h"

entity_t	r_worldentity;
vec3_t		r_modelorigin, r_entorigin;

int			r_visframecount;			// bumped when going to a new PVS
int			r_framecount;				// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys;

entity_t	*currententity;
int			currentvisedict;			// number of current visedict so we can reference the above array

gltexture_t particletexture;
int			playertextures[16];			// up to 16 color translated skins

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

//
// frame timers
//
float	r_frametime;
float	r_oldframetime = 0;

//
// screen size info
//
refdef_t	r_refdef;
mleaf_t		*r_viewleaf, *r_oldviewleaf;
texture_t	*r_notexture_mip;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value

// save out some space for verts to be reused for shadow drawing.
// these are kept statically here rather than in the entity_t struct so that they
// don't pollute instances of that struct where they're not actually needed...
// we need to store every vert in the model for this operation.
// we save all 3 verts as well in case we ever need to reuse the list
float shadowverts[MAX_VISEDICTS][MAXALIASVERTS][3];

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawentities = {"r_drawentities","1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_speeds = {"r_speeds","0"};
cvar_t	r_lightmap = {"r_lightmap","0"};
cvar_t  r_fullbright = {"r_fullbright", "0"};
cvar_t	r_shadows = {"r_shadows", "0.5", true};
cvar_t	r_wateralpha = {"r_wateralpha","1"};
cvar_t	r_dynamic = {"r_dynamic","1"};
cvar_t	r_novis = {"r_novis","0"};

cvar_t	gl_finish = {"gl_finish","0"};
cvar_t  gl_lockpvs = {"gl_lockpvs", "0"};
cvar_t	gl_clear = {"gl_clear","0"};
cvar_t	gl_cull = {"gl_cull","1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels","1"};
cvar_t	gl_affinemodels = {"gl_affinemodels","0"};
cvar_t	gl_polyblend = {"gl_polyblend","1"};
cvar_t	gl_playermip = {"gl_playermip","0"};
cvar_t	gl_nocolors = {"gl_nocolors","0"};
cvar_t	gl_keeptjunctions = {"gl_keeptjunctions","0"};
cvar_t	gl_reporttjunctions = {"gl_reporttjunctions","0"};
cvar_t	gl_doubleeyes = {"gl_doubleeys", "1"};
cvar_t  gl_notrans = {"gl_notrans", "0"};

/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	int		i;

	for (i = 0 ; i < 4 ; i++)
	{
		if (BoxOnPlaneSide (mins, maxs, &frustum[i]) == 2)
		{
			return true;
		}
	}
	return false;
}

/*
=============
R_CullSphere

Returns true if the sphere is completely outside the frustum
=============
*/
qboolean R_CullSphere (vec3_t centre, float radius)
{
	int		i;

	for (i = 0; i < 4; i++)
	{
		if (PlaneDiff(centre, &frustum[i]) <= -radius)
		{
			return true;
		}
	}
	return false;
}

/*
=================
R_BoxContainsPoint
=================
*/
qboolean R_BoxContainsPoint (vec3_t p, vec3_t mins, vec3_t maxs, float fudge)
{
	if (p[0] + fudge > maxs[0]) return false;
	if (p[1] + fudge > maxs[1]) return false;
	if (p[2] + fudge > maxs[2]) return false;

	if (p[0] - fudge < mins[0]) return false;
	if (p[1] - fudge < mins[1]) return false;
	if (p[2] - fudge < mins[2]) return false;

	return true;
}

/*
=================
R_RotateForEntity
=================
*/
void R_RotateForEntity (entity_t *e, qboolean blendLerp)
{
	float	timepassed;
	float	blend;
	vec3_t	d;
	int		i;

	if (blendLerp)
	{
		timepassed = realtime - e->translateStartTime;

		if (e->translateStartTime == 0 || timepassed > 1)
		{
			e->translateStartTime = realtime;
			VectorCopy (e->origin, e->origin1);
			VectorCopy (e->origin, e->origin2);
		}

		if (!VectorCompare (e->origin, e->origin2))
		{
			e->translateStartTime = realtime;
			VectorCopy (e->origin2, e->origin1);
			VectorCopy (e->origin,  e->origin2);
			blend = 0;
		}
		else
		{
			blend = timepassed / 0.1;

			if (cl.paused || blend > 1) blend = 1;
		}
		VectorSubtract (e->origin2, e->origin1, d);

		glTranslatef (e->origin1[0] + (blend * d[0]), e->origin1[1] + (blend * d[1]), e->origin1[2] + (blend * d[2]));

		// orientation interpolation (Euler angles, yuck!)
		timepassed = realtime - e->rotateStartTime;

		if (e->rotateStartTime == 0 || timepassed > 1)
		{
			e->rotateStartTime = realtime;
			VectorCopy (e->angles, e->angles1);
			VectorCopy (e->angles, e->angles2);
		}

		if (!VectorCompare (e->angles, e->angles2))
		{
			e->rotateStartTime = realtime;
			VectorCopy (e->angles2, e->angles1);
			VectorCopy (e->angles,  e->angles2);
			blend = 0;
		}
		else
		{
			blend = timepassed / 0.1;

			if (cl.paused || blend > 1) blend = 1;
		}
		VectorSubtract (e->angles2, e->angles1, d);

		// always interpolate along the shortest path
		for (i = 0; i < 3; i++)
		{
			if (d[i] > 180)
			{
				d[i] -= 360;
			}
			else if (d[i] < -180)
			{
				d[i] += 360;
			}
		}
		glRotatef (e->angles1[1] + ( blend * d[1]),  0, 0, 1);

		// store the matrix out for shadow drawing (saves us having to do all those
		// transforms again).  save it out *before* doing the scaling or the rest of
		// the rotation, cos that fucks with our shadow positioning
		if (e->model->type == mod_alias)
		{
			glGetFloatv (GL_MODELVIEW_MATRIX, e->mvievmatrix);
		}
		glRotatef (-e->angles1[0] + (-blend * d[0]),  0, 1, 0);
		glRotatef ( e->angles1[2] + ( blend * d[2]),  1, 0, 0);
	}
	else
	{
		glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);
		glRotatef (e->angles[1],  0, 0, 1);

		// store the matrix out for shadow drawing (saves us having to do all those
		// transforms again).  save it out *before* doing the scaling or the rest of
		// the rotation, cos that fucks with our shadow positioning
		if (e->model->type == mod_alias)
		{
			glGetFloatv (GL_MODELVIEW_MATRIX, e->mvievmatrix);
		}
		glRotatef (-e->angles[0],  0, 1, 0);
		glRotatef (e->angles[2],  1, 0, 0);
	}
}

/*
=============================================================

  SPRITE MODELS

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *ent)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes;
	float			lastmsg;
	float			*pintervals, fullinterval, targettime, time;

	// get the data
	psprite = (msprite_t *) ent->model->extradata;

	// terminate lastmessage properly.
	memset (&lastmsg, 0, sizeof(float));

	if ((ent->frame >= psprite->numframes) || (ent->frame < 0))
	{
		if (COM_Timeout (&lastmsg, 2) && (ent->frame == ent->model->maxframe || ent->frame < 0) || ent->frame > ent->model->maxframe)
		{
			if (ent->frame > ent->model->maxframe)
			{
				ent->model->maxframe = ent->frame; // Save peak
			}
			Con_Printf ("R_GetSpriteFrame: no such frame %d (%d frames) in %s\n", ent->frame, psprite->numframes, ent->model->name);
		}
		ent->frame = 0;
	}

	if (psprite->frames[ent->frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[ent->frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[ent->frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];
		time = cl.time + ent->syncbase;

		// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
		// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)	break;
		}
		pspriteframe = pspritegroup->frames[i];
	}
	return pspriteframe;
}

/*
=================
R_DrawSpriteModel
=================
*/
void R_DrawSpriteModel (entity_t *ent)
{
	vec3_t			point;
	mspriteframe_t	*frame;
	msprite_t		*psprite;
	vec3_t			up, right;
	vec3_t			v_forward, v_right, v_up;
	vec3_t			fixed_origin;
	vec3_t			spritecolor;
	vec3_t			temp;

	// get the data
	psprite = (msprite_t *) ent->model->extradata;

	switch (psprite->replace)
	{
	case P_BUBBLE:
		R_ParticleSprite (244, 8, ent->origin, P_BUBBLE);
		return;

	case P_FLARE2:
		R_ParticleSprite (108, 64, ent->origin, P_FLARE2);
		return;

	case P_BLAST:
		// this doesn't call a replacement function but is flagged all the same
		// so that it is not drawn as a sprite.
		return;

	case P_RAIN:
		R_ParticleSprite (47, 6, ent->origin, P_RAIN);
		return;

	case P_NONE:
	default:
		// do nothing - here we just fall through to the sprite drawer
		break;
	}
	frame = R_GetSpriteFrame (ent);

	VectorCopy(ent->origin, fixed_origin);

	if (psprite->type == SPR_ORIENTED)
	{
		// bullet marks on walls
		AngleVectors (ent->angles, v_forward, v_right, v_up);
		VectorScale(v_forward, -2, temp);
		VectorAdd(temp, fixed_origin, fixed_origin);
		VectorCopy (v_up, up);
		VectorCopy (v_right, right);

		// calculate a dynamic light for sprite
		R_LightPoint(ent->origin, spritecolor);

		// just some default values
		spritecolor[0] = spritecolor[0] * 0.00390625f;
		spritecolor[1] = spritecolor[1] * 0.00390625f;
		spritecolor[2] = spritecolor[2] * 0.00390625f;
	}
	else if (psprite->type == SPR_FACING_UPRIGHT)
	{
		up[0] = 0;
		up[1] = 0;
		up[2] = 1;

		right[0] = ent->origin[1] - r_origin[1];
		right[1] = -(ent->origin[0] - r_origin[0]);
		right[2] = 0;

		Normalize (right);
	}
	else if (psprite->type == SPR_VP_PARALLEL_UPRIGHT)
	{
		up[0] = 0;
		up[1] = 0;
		up[2] = 1;

		VectorCopy (vright, right);
	}
	else
	{
		// normal sprite
		VectorCopy (vup, up);
		VectorCopy (vright, right);
	}

	// dynamically lit sprites woo boy
	if (psprite->type == SPR_ORIENTED)
	{
		glColor3f (spritecolor[0] * 1, spritecolor[1] * 1, spritecolor[2] * 1);
	}
	else
	{
		glColor3f(1, 1, 1);
	}
	glBindTexture (GL_TEXTURE_2D, frame->gl_texture->gltexnum);

	glEnable (GL_ALPHA_TEST);
	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (fixed_origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (fixed_origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (fixed_origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (fixed_origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glEnd ();

	glDisable (GL_ALPHA_TEST);
}

/*
=============================================================

  ALIAS MODELS

=============================================================
*/


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] =
{
#include "anorms.h"
};

vec3_t	shadevector;
float	shadelight[3], ambientlight;

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
	;

float	*shadedots1 = r_avertexnormal_dots[0];
float	*shadedots2 = r_avertexnormal_dots[0];

int		lastposenum;
int		lastposenum0;

float	lightLerpOffset;


/*
=============
GL_DrawAliasFrame

we can't cache alias model verts because they can change many times, even in the same frame, so instead
we calculate on the fly.  we might increase our struct to allow for this, or add a big dirty static
array to the pheader at some stage.
=============
*/
void GL_DrawAliasFrame (aliashdr_t *paliashdr, int pose1, int pose2, byte constantlight, float blend, qboolean pass2)
{
	float 		l = 0;
	float		ldiff;
	float		lquant[2];
	drawvertx_t	*verts1;
	drawvertx_t	*verts2;
	int			*order;
	int			count;
	vec3_t		diff;
	int			vcount;
	GLenum		TRIANGLE_DRAW_MODE;
	int			totalvcount = 0;

	lastposenum0 = pose1;
	lastposenum = pose2;

	verts1 = (drawvertx_t *) ((byte *) paliashdr + paliashdr->posedata);
	verts2 = verts1;

	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	order = (int *) ((byte *) paliashdr + paliashdr->commands);

	if (constantlight)
	{
		l = (float) constantlight / 255.0;
	}

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		if (count < 0)
		{
			count = -count;
			TRIANGLE_DRAW_MODE = GL_TRIANGLE_FAN;
		}
		else TRIANGLE_DRAW_MODE = GL_TRIANGLE_STRIP;

		// the vertex counter starts at 0 for each fan/strip group
		vcount = 0;

		do
		{
			// texture s and t
			aliasverts[vcount].st[0] = ((float *) order)[0];
			aliasverts[vcount].st[1] = ((float *) order)[1];

			if (!constantlight)
			{
				lquant[0] = (shadedots1[verts1->lightnormalindex] + (blend * (shadedots1[verts2->lightnormalindex] - shadedots1[verts1->lightnormalindex])));
				lquant[1] = (shadedots2[verts1->lightnormalindex] + (blend * (shadedots2[verts2->lightnormalindex] - shadedots2[verts1->lightnormalindex])));

				if (lquant[0] != lquant[1])
				{
					if (lquant[0] > lquant[1])
					{
						ldiff = lquant[0] - lquant[1];
						ldiff *= lightLerpOffset;
						l = lquant[0] - ldiff;
					}
					else
					{
						ldiff = lquant[1] - lquant[0];
						ldiff *= lightLerpOffset;
						l = lquant[0] + ldiff;
					}
				}
				else
				{
					l = lquant[0];
				}

				// square it cos it looks pretty...
				l *= l;

				// only set translucency if the mdl is lit
				aliasverts[vcount].rgba[0] = l * shadelight[0];
				aliasverts[vcount].rgba[1] = l * shadelight[1];
				aliasverts[vcount].rgba[2] = l * shadelight[2];
				aliasverts[vcount].rgba[3] = paliashdr->translevel;
			}
			else
			{
				aliasverts[vcount].rgba[0] = l;
				aliasverts[vcount].rgba[1] = l;
				aliasverts[vcount].rgba[2] = l;
				aliasverts[vcount].rgba[3] = 1.0f;
			}

			// vertex co-ords
			VectorSubtract (verts2->v, verts1->v, diff);

			// set up the base verts for drawing
			aliasverts[vcount].v[0] = verts1->v[0] + (blend * diff[0]);
			aliasverts[vcount].v[1] = verts1->v[1] + (blend * diff[1]);
			aliasverts[vcount].v[2] = verts1->v[2] + (blend * diff[2]);

			if (!pass2)
			{
				// save out the shadow verts on pass 1 only because subsequent passes may not be lerped!!!
				shadowverts[currentvisedict][totalvcount][0] = aliasverts[vcount].v[0];
				shadowverts[currentvisedict][totalvcount][1] = aliasverts[vcount].v[1];
				shadowverts[currentvisedict][totalvcount][2] = aliasverts[vcount].v[2];
			}

			// increment the lists
			order += 2;
			verts1++;
			verts2++;

			// and the vertex counter
			vcount++;
			totalvcount++;
		}
		while (--count);

		// draw from index 0 for each tri fan or strip
		glDrawArrays (TRIANGLE_DRAW_MODE, 0, vcount);
	}
}


/*
=================
R_DrawAliasModelShadow

here's how it SHOULD be done.  heh heh heh.

we stored out the verts for the current frame when drawing the model, so we can just use
them again here without having to run through frame setup and blend calculation.  instant
interpolated shadows with half the CPU hit!!!  woo-oooooo-oooooooo-oooo!!!

everything is batched up and drawn in pass of its own so as to be able to do it with only
one set of state changes.  the modelview matrix for each alias model is saved out to the
entity structure so we can just load it straight in rather than sending it through
r_rotateforentity (which can be expensive if there's loads of sins and cos's).
=================
*/
void R_DrawAliasModelShadow (entity_t *ent)
{
	int					count;
	int					vcount;
	GLenum				TRIANGLE_DRAW_MODE;
	float				height;
	float				s1, c1;
	aliashdr_t			*paliashdr;
	int					*order;
	int					totalvcount = 0;

    // culled during previous pass
    if (ent->culled) return;

	// no lightplane so don't even bother
	if (!ent->lightplane) return;

	// load the models modelview matrix directly
	// see how this is cleaner than running through the transforms
	glLoadMatrixf (ent->mvievmatrix);

	// use the DIST_EPSILON from world.c (moved into quakedef.h)
	height = DIST_EPSILON - (ent->origin[2] - ent->lightspot[2]);

	s1 = sin (ent->angles[1] * DIV180MULTPI);
	c1 = cos (ent->angles[1] * DIV180MULTPI);

	paliashdr = (aliashdr_t *) ent->model->extradata;
	order = (int *) ((byte *) paliashdr + paliashdr->commands);

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		if (count < 0)
		{
			count = -count;
			TRIANGLE_DRAW_MODE = GL_TRIANGLE_FAN;
		}
		else TRIANGLE_DRAW_MODE = GL_TRIANGLE_STRIP;

		// the vertex counter starts at 0 for each fan/strip group
		vcount = 0;

		do
		{
			// restore the shadow verts for indexes 0 and 1
			aliasverts[vcount].v[0] = shadowverts[currentvisedict][totalvcount][0];
			aliasverts[vcount].v[1] = shadowverts[currentvisedict][totalvcount][1];

			// collapse the verts on a vertical plane to get index 2
			// i'm not angling the verts cos i want the shadow directly under the model, but i do the
			// aliasVerts[2] thing anyway cos (a) it keeps the value of aliasverts (lcase v) consistent,
			// and (b) it's easier for someone to mod my engine if they want this way.
			aliasverts[vcount].v[2] = height;

			// do the sloped surface thing.  this isn't perfect - e.g. where a sloped surface meets a regular
			// one, it gets fucked.  may have to project each point in each vert downwards
			aliasverts[vcount].v[2] += ((aliasverts[vcount].v[1] * (s1 * ent->lightplane->normal[0])) -
										(aliasverts[vcount].v[0] * (c1 * ent->lightplane->normal[0])) -
										(aliasverts[vcount].v[0] * (s1 * ent->lightplane->normal[1])) -
										(aliasverts[vcount].v[1] * (c1 * ent->lightplane->normal[1]))) +
									   ((1.0 - ent->lightplane->normal[2]) * 20);

			// increment the list
			order += 2;

			// and the vertex counter
			vcount++;
			totalvcount++;
		}
		while (--count);

		// draw from index 0 for each tri fan or strip
		glDrawArrays (TRIANGLE_DRAW_MODE, 0, vcount);
	}
}

/*
=================
R_SetupAliasFrame

contains full interpolation code - it doesn't make much difference having it, cos
even if we don't interpolate, nothing adverse is done here...
=================
*/
void R_SetupAliasFrame (entity_t *ent, aliashdr_t *paliashdr)
{
	int			pose, numposes;
	int			texture, lumatexture, normtexture;
	float		lastmsg;
	float		blend;
	int			anim;
	model_t		*mod;
	qboolean    framelerp;

	mod = ent->model;

	// terminate lastmessage properly.
	memset (&lastmsg, 0, sizeof(float));

	// check frames
	if (ent->frame >= paliashdr->numframes || ent->frame < 0)
	{
		if (COM_Timeout(&lastmsg, 2) && (ent->frame == mod->maxframe || ent->frame < 0) || ent->frame > mod->maxframe)
		{
			if (ent->frame > ent->model->maxframe)
			{
				mod->maxframe = ent->frame; // Save peak
			}
			Con_DPrintf ("%s: no such frame ", "R_SetupAliasFrame");

			// Single frame?
			if (paliashdr->frames[0].name[0])
			{
				Con_DPrintf ("%d ('%s', %d frames)", ent->frame, paliashdr->frames[0].name, paliashdr->numframes);
			}
			else
			{
				Con_DPrintf ("group %d (%d groups)", ent->frame, paliashdr->numframes);
			}
			Con_DPrintf (" in %s\n", mod->name);
		}
		ent->frame = 0;
	}

	// get poses
	pose = paliashdr->frames[ent->frame].firstpose;
	numposes = paliashdr->frames[ent->frame].numposes;

	if (numposes > 1)
	{
		ent->frameInterval = paliashdr->frames[ent->frame].interval;
		pose += (int)(realtime / ent->frameInterval) % numposes;
	}
	else
	{
		ent->frameInterval = 0.1f;
	}

	// conditions for turning lerping off
	framelerp = ((ent->pose2 != pose) || (paliashdr->numposes == 1) || (ent == &cl.viewent && ent->frame == 0) || (r_framecount < 10)) ? false : true;

	// correctly switch interpolation off
	// (viewent can switch mdls but keep the same pose, and therefore have invalid data cached in the poses)
	if (!framelerp)
	{
		ent->frameStartTime = realtime;
		ent->pose1 = ent->pose2;
		ent->pose2 = pose;
		blend = 0.0f;
	}
	else
	{
		blend = (realtime - ent->frameStartTime) / ent->frameInterval;
	}

	// don't let blend pass 1
	if (cl.paused || blend > 1.0f)
	{
		blend = 1.0f;
	}

	// get the current animation
	anim = (int) (cl.time * 10) & 3;

	// check invalid skins
	if ((ent->skinnum >= paliashdr->numskins) || (ent->skinnum < 0))
	{
		Con_DPrintf ("R_DrawAliasModel: no such skin # %d\n", ent->skinnum);
		ent->skinnum = 0;
	}

	// texture pointers
	texture = paliashdr->gl_texture[ent->skinnum][anim]->gltexnum;
    normtexture = paliashdr->gl_texture[ent->skinnum][anim]->normtexnum;
	lumatexture = paliashdr->gl_texture[ent->skinnum][anim]->lumatexnum;

	// bind the base texture
	glBindTexture (GL_TEXTURE_2D, texture);

	// player texture crap...
	// we can't dynamically colormap textures, so they are cached
	// seperately for the players.  Heads are just uncolored.
	if (ent->colormap != vid.colormap && !gl_nocolors.value)
	{
		int i = ent - cl_entities;

		if (i >= 1 && i <= cl.maxclients)
		{
			glBindTexture (GL_TEXTURE_2D, playertextures[i - 1]);
		}
	}

	// store the drawlist in the alias model header
	GL_DrawAliasFrame (paliashdr, ent->pose1, ent->pose2, 0, blend, false);

#if 0 // will eventually be for alias model bumpmapping
    if (normtexture)
    {
        // proper blend mode
        glBlendFunc (GL_DST_COLOR, GL_ZERO);
        glEnable (GL_BLEND);

        // bind the norm texture
        glBindTexture (GL_TEXTURE_2D, normtexture);

        // store the drawlist in the alias model header
        GL_DrawAliasFrame (paliashdr, ent->pose1, ent->pose2, 0, blend, false);

        // blend off
        glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glDisable (GL_BLEND);
    }
#endif

	// luma luma luma
	if (lumatexture)
	{
		int temppose1;
		int temppose2;

		// take down fog before drawing if it's underwater...
		if (ent->modelleaf->volumetexture)
		{
			glDisable (GL_FOG);
		}

		// disable blending on the view model to prevent muzzle flash weirdness
		if (ent == &cl.viewent)
		{
			// correctly switch interpolation off
			temppose1 = ent->pose2;
			temppose2 = pose;
			blend = 0;
		}
		else
		{
			temppose1 = ent->pose1;
			temppose2 = ent->pose2;
		}

		// bind the luma texture (we assume this won't change for player skins)
		glBindTexture (GL_TEXTURE_2D, lumatexture);

		// draw fully lit
		GL_DrawAliasFrame (paliashdr, temppose1, temppose2, 255, blend, true);

		// bring fog back up after drawing if it's underwater
		if (ent->modelleaf->volumetexture)
		{
			glEnable (GL_FOG);
		}
	}
}

/*
=================
R_DrawAliasModel
=================
*/
void R_DrawAliasModel (entity_t *ent, qboolean viewalias)
{
	int			i;
	aliashdr_t	*paliashdr;
	float		ang_ceil, ang_floor;

    // culled during previous pass
    if (ent->culled) return;

	// copy the different origins
	VectorCopy (ent->origin, r_entorigin);
	VectorSubtract (r_origin, r_entorigin, r_modelorigin);

	// set to NULL initially - if R_LightPoint fails then we can do a test
	ent->lightplane = NULL;
	lightplane = NULL;

	// always do this cos we need lightspot for shadows
	R_LightPoint (currententity->origin, shadelight);

	// save out lightspot and lightplane for shadows
	ent->lightspot[0] = lightspot[0];
	ent->lightspot[1] = lightspot[1];
	ent->lightspot[2] = lightspot[2];
	ent->lightplane = lightplane;

	// rotating objects (i.e. pick-ups) will always have some constant light in order to
	// make them noticeable... (make this pulse or flicker...)
	if (ent->model->flags & EF_ROTATE)
	{
		shadelight[0] = 192;
		shadelight[1] = 192;
		shadelight[2] = 192;
	}
	else
	{
		// always give the gun some light
		if (ent == &cl.viewent)
		{
			if (shadelight[0] < 16) shadelight[0] = 16;
			if (shadelight[1] < 16) shadelight[1] = 16;
			if (shadelight[2] < 16) shadelight[2] = 16;
		}

		// never allow players to go totally black
		if ((ent - cl_entities) >= 1 && (ent - cl_entities) <= cl.maxclients)
		{
			if (shadelight[0] < 8) shadelight[0] = 8;
			if (shadelight[1] < 8) shadelight[1] = 8;
			if (shadelight[2] < 8) shadelight[2] = 8;
		}

		// clamp lighting so it doesn't overbright as much
		if (shadelight[0] > 192) shadelight[0] = 192;
		if (shadelight[1] > 192) shadelight[1] = 192;
		if (shadelight[2] > 192) shadelight[2] = 192;
	}

	// light interpolation
	lightLerpOffset = (ent->angles[1] + ent->angles[0]) * (SHADEDOT_QUANT / 360.0);

	ang_ceil = ceil (lightLerpOffset);
	ang_floor = floor (lightLerpOffset);

	lightLerpOffset = ang_ceil - lightLerpOffset;

	shadedots1 = r_avertexnormal_dots[(int) ang_ceil & (SHADEDOT_QUANT - 1)];
	shadedots2 = r_avertexnormal_dots[(int) ang_floor & (SHADEDOT_QUANT - 1)];

	// set to a 0 to 1 scale, and reduce a little
	shadelight[0] = shadelight[0] / 224.0f;
	shadelight[1] = shadelight[1] / 224.0f;
	shadelight[2] = shadelight[2] / 224.0f;

	// interpolate and store for the next frame
	for (i = 0; i < 3; i++)
	{
		if (ent->lastShadeLight[i])
		{
			shadelight[i] = (shadelight[i] + ent->lastShadeLight[i]) / 2;
		}
		ent->lastShadeLight[i] = shadelight[i];
	}

	// locate the proper data
	paliashdr = (aliashdr_t *) ent->model->extradata;

	// draw all the triangles
	glPushMatrix ();

	R_RotateForEntity (ent, !viewalias);

	R_SetupAliasFrame (ent, paliashdr);

	glPopMatrix ();

	// add particle sprites for fx
	if (paliashdr->glow)
	{
		vec3_t			baseorigin;
		float			baseradius = paliashdr->radius;
		pdrawstyle_t	basetype = P_FLARE2;

		// take down fog before drawing if it's underwater...
		if (ent->modelleaf->volumetexture)
		{
			glDisable (GL_FOG);
		}
		VectorCopy (ent->origin, baseorigin);

		// set effects
		if (paliashdr->torch)
		{
			baseorigin[2] += 8;
		}

		if (paliashdr->flicker)
		{
			baseradius += (float) (Q_rand () % 6);
		}

		if (paliashdr->missile)
		{
			// move the glow behind the rocket - fixme - do a fast sin/cos lookup table for this...
			baseorigin[0] += cos (ent->angles[1] * DIV180MULTPI) * (-20.0f);
			baseorigin[1] += sin (ent->angles[1] * DIV180MULTPI) * (-20.0f);
			baseorigin[2] += sin (ent->angles[0] * DIV180MULTPI) * (-20.0f);

			basetype = P_FLARE1;
		}

		if (paliashdr->shambler)
		{
			// shambler magic hack cos the origin of the s_light mdl used for this is
			// in the shamblers groin - which, though interesting, is just not right...
			baseorigin[2] += 75.0;

			basetype = P_FLARE1;
		}

		if (paliashdr->powerup)
		{
			// simple static glow
			baseorigin[2] += 16;

			basetype = P_FLARE1;
		}
		R_ParticleSprite (paliashdr->colour, baseradius, baseorigin, basetype);

		// bring fog back up after drawing if it's underwater
		if (ent->modelleaf->volumetexture)
		{
			glEnable (GL_FOG);
		}
	}
}


//==================================================================================

/*
=============
R_DepthSortAliasEntities
=============
*/
static int R_DepthSortAliasEntities (const void *a, const void *b)
{
	static int	offset;
	entity_t	*enta = *((entity_t **) a);
	entity_t	*entb = *((entity_t **) b);

	// sort back to front
	if (enta->distance > entb->distance)
	{
		offset = -1;
	}
	else if (enta->distance < entb->distance)
	{
		offset = 1;
	}
	else
	{
		offset = 0;
	}
	return offset;
}

/*
================
R_TransSortAliasEntities
================
*/
void R_TransSortAliasEntities (void)
{
	int	i;

	// if there's only one then the list is already sorted!
	if (cl_numvisedicts > 1)
	{
		// evaluate distances
		for (i = 0; i < cl_numvisedicts; i++)
		{
			entity_t *ent = cl_visedicts[i];

			// set distance from viewer - no need to sqrt them as the order will be the same
			ent->distance = (ent->origin[0] - r_origin[0]) * (ent->origin[0] - r_origin[0]) +
							(ent->origin[1] - r_origin[1]) * (ent->origin[1] - r_origin[1]) +
							(ent->origin[2] - r_origin[2]) * (ent->origin[2] - r_origin[2]);
		}

		if (cl_numvisedicts == 2)
		{
			// trivial case - 2 entities
			if (cl_visedicts[0]->distance < cl_visedicts[1]->distance)
			{
				entity_t *tmpent = cl_visedicts[1];

				// reorder correctly
				cl_visedicts[1] = cl_visedicts[0];
				cl_visedicts[0] = tmpent;
			}
		}
		else
		{
			// general case - depth sort the transedicts from back to front
			qsort ((void *) cl_visedicts, cl_numvisedicts, sizeof (entity_t *), R_DepthSortAliasEntities);
		}
	}
}

/*
=============
R_SetupAliasDraw

at some point this will get modified to also
include the code for bumpmapping models.
=============
*/
void R_SetupAliasDraw (void)
{
	// set up vertex arrays
	glEnableClientState (GL_VERTEX_ARRAY);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glEnableClientState (GL_COLOR_ARRAY);

    // tmu0 is active here
	glClientActiveTexture (GL_TEXTURE0);

	// set the vertex array pointers
	glVertexPointer (3, GL_FLOAT, sizeof (aliasverts_t), &((float *) &aliasverts)[0]);
	glTexCoordPointer (2, GL_FLOAT, sizeof (aliasverts_t), &((float *) &aliasverts)[3]);
	glColorPointer (4, GL_FLOAT, sizeof (aliasverts_t), &((float *) &aliasverts)[5]);

	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);	// for translucency
	glEnable (GL_BLEND);	// for fullbrights

	// batch these state changes up here
	glShadeModel (GL_SMOOTH);

	if (gl_affinemodels.value)
	{
		glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
	}
}

/*
=============
R_ShutdownAliasDraw

at some point this will get modified to also
include the code for bumpmapping models.
=============
*/
void R_ShutdownAliasDraw (void)
{
	glShadeModel (GL_FLAT);

	if (gl_affinemodels.value)
	{
		glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	}
	glDisable (GL_BLEND);
	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glDisableClientState (GL_VERTEX_ARRAY);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_COLOR_ARRAY);
}

/*
=============
R_SetupAliasShadowDraw
=============
*/
void R_SetupAliasShadowDraw (void)
{
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (aliasverts_t), &((float *) &aliasverts)[0]);
	glDisable (GL_TEXTURE_2D);

	// allow alpha to be set by the value of r_shadows
	glColor4f (0, 0, 0, r_shadows.value > 1 ? 1 : r_shadows.value);

	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);

	glEnable (GL_STENCIL_TEST);
	glStencilFunc (GL_EQUAL, 1, 2);
	glStencilOp (GL_KEEP, GL_KEEP, GL_INCR);
}

/*
=============
R_ShutdownAliasShadowDraw
=============
*/
void R_ShutdownAliasShadowDraw (void)
{
	// done - restore state
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	glDisable (GL_STENCIL_TEST);

	glDisableClientState (GL_VERTEX_ARRAY);
	glEnable (GL_TEXTURE_2D);
	glColor4f (1, 1, 1, 1);

	// go back to the world matrix (cant use a pointer here or our viewmodel gets fucked)
	glLoadMatrixf (cl_entities[0].mvievmatrix);
}

/*
=============
R_SetupAliasUnderwaterDraw
=============
*/
void R_SetupAliasUnderwaterDraw (void)
{
	// now do caustics
	if (r_caustictexture.gltexnum)
	{
		// now for the caustic texture.  because these models normally move (with the exception of
		// pickups and the viewmodel) we can afford to be a little coarser than we were for the world
		glActiveTexture (GL_TEXTURE1);
		glEnable (GL_TEXTURE_2D);
		glBindTexture (GL_TEXTURE_2D, r_caustictexture.gltexnum);

		// multiply and rescale colours to keep the original intensity levels
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
		glTexEnvf (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
		glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE);
		glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
		glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_PREVIOUS);
		glTexEnvf (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2);

		// don't bother with any prettiness in here, like warping s and t in their
		// own verts.  this will give more than adequate results for mdls...
		glMatrixMode (GL_TEXTURE);
		glLoadIdentity ();
		glScalef (0.5, 0.5, 1);
		glRotatef (realtime * 10, 1, 0, 0);
		glRotatef (realtime * 10, 0, 1, 0);
		glMatrixMode (GL_MODELVIEW);

		// reuse the same s and t as the skin
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (aliasverts_t), &((float *) &aliasverts)[3]);

		// back to tmu 0 for the skins
		glActiveTexture (GL_TEXTURE0);
		glClientActiveTexture (GL_TEXTURE0);
	}
}

/*
=============
R_ShutdownAliasUnderwaterDraw
=============
*/
void R_ShutdownAliasUnderwaterDraw (void)
{
	// now shutdown caustics
	if (r_caustictexture.gltexnum)
	{
		// tmu madness (god, this stuff can get wearying after a while)
		glActiveTexture (GL_TEXTURE1);

		// this is needed too otherwise it totally craps out with other combines on GL_TEXTURE1...
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1);
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glDisable (GL_TEXTURE_2D);

		glMatrixMode (GL_TEXTURE);
		glLoadIdentity ();
		glMatrixMode (GL_MODELVIEW);

		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);

		glClientActiveTexture (GL_TEXTURE0);
		glActiveTexture (GL_TEXTURE0);
	}
}

/*
=============
R_SetupVolumeEffect
=============
*/
void R_SetupVolumeEffect (entity_t *ent)
{
    // this is underwater we dont want to fuckup fog above water.
	if (ent->modelleaf->volumetexture)
    {
        // turn on fog.  we'll do this in opengl rather than software emulation cos
        // there's too many verts in a model and the extra pass would crucify us.
        // note that we don't set colour at this stage cos it's different for each mdl
        glFogf (GL_FOG_MODE, GL_LINEAR);

        // this looks like a Thing That Should Not Be, but it's actually what we want to do, to
        // ensure a minimum amount of fogging.  this will not look good when we're in lava or
        // slime, but by that point we've probably got all sorts of other things on our mind...
        glFogf (GL_FOG_START, -500);
        glFogf (GL_FOG_END, 1024);
        glFogf (GL_FOG_DENSITY, 0.2f);

        // fog colour is potentially different for each mdl, so we must
        // set it seperately for each.
        glFogfv (GL_FOG_COLOR, (const GLfloat *)ent->modelleaf->volumetexture->gl_texture->rgba);

        glHint (GL_FOG_HINT, GL_NICEST);
        glEnable (GL_FOG);
    }
}

/*
=============
R_ShutDownVolumeEffect
=============
*/
void R_ShutDownVolumeEffect (entity_t *ent)
{
	// this is underwater we dont want to fuckup fog above water
	if (ent->modelleaf->volumetexture)
    {
        // turn off fog
        glDisable (GL_FOG);
    }
}

/*
=============
R_DrawAliasEntities
=============
*/
void R_DrawAliasEntities (void)
{
	int			i;
	qboolean	pickupalias = false;
	qboolean	pickupshadows = false;

	if (!r_drawentities.value) return;

    // depth sort alias entities
    R_TransSortAliasEntities ();

	// add the state we need for alias models
	R_SetupAliasDraw ();

	// draw sprites seperately, because of alpha blending
	for (i = 0; i < cl_numvisedicts; i++)
	{
		// store it in currententity
		currententity = cl_visedicts[i];
		currentvisedict = i;

		// entity has no model!!!
		if (!currententity->model) continue;

		// not an alias model
		if (currententity->model->type != mod_alias) continue;

        // get entity origin
        VectorAdd (currententity->origin, currententity->model->mins, currententity->mins);
        VectorAdd (currententity->origin, currententity->model->maxs, currententity->maxs);

        // do a leaf lookup to set some defaults for caustics / fog.
        // need to do this before any culling is done as else we wont have the nessesary data.
        currententity->modelleaf = R_LeafForEntity (currententity->mins, currententity->maxs);

		// adjust the bounding box if the model has been rotated
		if (currententity->angles[0] || currententity->angles[1] || currententity->angles[2])
		{
			// mark as culled if the sphere is outside the frustum so that it wont have to be tested again
			currententity->culled = R_CullSphere (currententity->origin, currententity->model->radius);
		}
		else
		{
			// mark as culled if the box is outside the frustum so that it wont have to be tested again
			currententity->culled = R_CullBox (currententity->mins, currententity->maxs);
		}

		// test for drawing
		if (!currententity->modelleaf->volumetexture)
		{
			// only draw alias mdls in solid and empty leafs during this pass
			R_DrawAliasModel (currententity, false);
		}
		else
		{
			// if not in solid or empty draw the pickup
			if (CONTENTMASK_LIQUID(r_viewleaf))
			{
                pickupalias = true;
			}
		}

		// dont bother shadowing the view entity
		if (currententity != &cl_entities[cl.viewentity])
		{
		    // if not underwater draw the pickup
		    if (CONTENTMASK_NOTLIQUID(r_viewleaf))
		    {
                pickupshadows = true;
		    }
		}
	}

	if (pickupalias)
	{
		// add the state we need for underwater effects
		R_SetupAliasUnderwaterDraw ();

		for (i = 0; i < cl_numvisedicts; i++)
		{
			// store it in currententity
			currententity = cl_visedicts[i];
			currentvisedict = i;

			// entity has no model!!!
			if (!currententity->model) continue;

			// not an alias model
			if (currententity->model->type != mod_alias) continue;

            // add water entity effect
            R_SetupVolumeEffect (currententity);

            // draw the model with the underwater effect
            R_DrawAliasModel (currententity, false);

            // take down water entity effect
            R_ShutDownVolumeEffect (currententity);
		}

		// remove underwater effects
		R_ShutdownAliasUnderwaterDraw ();
	}
	R_ShutdownAliasDraw ();

	if (pickupshadows)
	{
		// add the state we need for shadows
		R_SetupAliasShadowDraw ();

		for (i = 0; i < cl_numvisedicts; i++)
		{
			// store it in currententity
			currententity = cl_visedicts[i];
			currentvisedict = i;

			// entity has no model!!!
            if (!currententity->model) continue;

			// not an alias model
			if (currententity->model->type != mod_alias) continue;

            // draw the models shadow
            R_DrawAliasModelShadow (currententity);
		}
		R_ShutdownAliasShadowDraw ();
	}
}

/*
=============
R_DrawSpriteEntities

Allways drawn so we made it seperate.
=============
*/
void R_DrawSpriteEntities (void)
{
	int		i;

	if (!r_drawentities.value) return;

	// classic quake only has a handful of sprites: light globes, bubbles and explosions.
	// hipnotic added a few more, but they looked ghastly even back in the late 90s, so
	// we'll just pretend they never happened.  the main ones we want to deal with are
	// the globes and bubbles.  sprite replacement is handled in the drawing routine.
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		// store current entity, currentvisedict is skipped for sprites.
		currententity = cl_visedicts[i];
        currentvisedict = i;

		// entity has no model!!!
		if (!currententity->model) continue;

		// not an sprite model
		if (currententity->model->type != mod_sprite) continue;

        // draw sprite model
		R_DrawSpriteModel (currententity);
	}
}

/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{
	aliashdr_t	*viewmodheader;

	if (!r_drawviewmodel.value) return;
	if (chase_active.value) return;
	if (!r_drawentities.value) return;
	if (cl.stats[STAT_HEALTH] <= 0) return;

	currententity = &cl.viewent;

	// entity has no model!!!
	if (!currententity->model) return;

	cl.cutscene = true;

	// get a pointer for the translevel hack.
	viewmodheader = (aliashdr_t *) currententity->model->extradata;

	if (cl.items & IT_INVISIBILITY)
	{
		// set translucency if we have the ring
		viewmodheader->translevel = 0.4f;
	}

	// setup alias model drawing parameters
	R_SetupAliasDraw ();

	// hack the depth range to prevent view model from poking into walls
	// MH - and other alias models (torches, etc) on walls!!! (was 0.3)
	glDepthRange (gldepthmin, gldepthmin + 0.1 * (gldepthmax - gldepthmin));

	// don't bother culling the viewent cos it's always in the frustum.
	// (roughly bottom-center, i'd say, but don't ask me how i guessed that!!!)
	// this will always be the viewleaf (convenient!!!)
	currententity->modelleaf = r_viewleaf;

	// test for drawing
	if (currententity->modelleaf->volumetexture)
	{
		// setup underwater entity states
		R_SetupAliasUnderwaterDraw ();

		// add water entity effect
		R_SetupVolumeEffect (currententity);

		// draw it with the underwater effect
		R_DrawAliasModel (currententity, true);

		// take down water entity effect
		R_ShutDownVolumeEffect (currententity);

		// restore previous state
		R_ShutdownAliasUnderwaterDraw ();
	}
	else
	{
		// only draw alias mdls in solid and empty leafs during this pass
		R_DrawAliasModel (currententity, true);
	}

	// restore the depth range
	glDepthRange (gldepthmin, gldepthmax);

	// shutdown alias model drawing parameters
	R_ShutdownAliasDraw ();

	// switch off translucency after drawing
	if (cl.items & IT_INVISIBILITY)
	{
		viewmodheader->translevel = 1.0f;
	}
	cl.cutscene = false;
}


/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!gl_polyblend.value) return;
	if (!v_blend[3]) return;

	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

	glLoadIdentity ();

	glRotatef (-90,  1, 0, 0);	    // put Z going up
	glRotatef (90,  0, 0, 1);	    // put Z going up

	glColor4fv (v_blend);

	glBegin (GL_QUADS);
	glVertex3f (10, 100, 100);
	glVertex3f (10, -100, 100);
	glVertex3f (10, -100, -100);
	glVertex3f (10, 100, -100);
	glEnd ();

	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST);
}

/*
===============
R_SignbitsForPlane
===============
*/
int R_SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test
	bits = 0;

	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
		{
			bits |= 1<<j;
		}
	}
	return bits;
}

/*
===============
R_ExtractFrustum
===============
*/
void R_ExtractFrustum (void)
{
    int     i;
	float   proj[16];
	float   modl[16];
	float   clip[16];

	/* Get the current PROJECTION matrix from OpenGL */
	glGetFloatv( GL_PROJECTION_MATRIX, proj );

	/* Get the current MODELVIEW matrix from OpenGL */
	glGetFloatv( GL_MODELVIEW_MATRIX, modl );

	/* Combine the two matrices (multiply projection by modelview) */
	clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
	clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
	clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
	clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

	clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
	clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
	clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
	clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

	clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
	clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
	clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
	clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

	clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
	clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
	clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
	clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

    // right
    frustum[0].normal[0] = clip[3] - clip[0];
    frustum[0].normal[1] = clip[7] - clip[4];
    frustum[0].normal[2] = clip[11] - clip[8];

    // left
    frustum[1].normal[0] = clip[3] + clip[0];
    frustum[1].normal[1] = clip[7] + clip[4];
    frustum[1].normal[2] = clip[11] + clip[8];

    // bottom
    frustum[2].normal[0] = clip[3] + clip[1];
    frustum[2].normal[1] = clip[7] + clip[5];
    frustum[2].normal[2] = clip[11] + clip[9];

    // top
    frustum[3].normal[0] = clip[3] - clip[1];
    frustum[3].normal[1] = clip[7] - clip[5];
    frustum[3].normal[2] = clip[11] - clip[9];

    for (i = 0; i < 4; i++)
    {
        NormalizeFast (frustum[i].normal);

        frustum[i].type = PLANE_ANYZ;
        frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
        frustum[i].signbits = R_SignbitsForPlane (&frustum[i]);
    }
}

/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{
	R_AnimateLight ();

	// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);
	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

	// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	c_brush_polys = 0;
	c_alias_polys = 0;
}

/*
=============
GL_SetPerspective
=============
*/
void GL_SetPerspective (float fovx, float fovy)
{
#define NEARCLIP 4
    float left, right, bottom, top;
    float nudge = 1.0 - 1.0 / (1 << 23);

    right = NEARCLIP * tan (fovx * M_PI / 360.0);
    left = -right;

    top = NEARCLIP * tan (fovy * M_PI / 360.0);
    bottom = -top;

    cl_entities[0].pjectmatrix[0] = (2 * NEARCLIP) / (right - left);
    cl_entities[0].pjectmatrix[4] = 0;
    cl_entities[0].pjectmatrix[8] = (right + left) / (right - left);
    cl_entities[0].pjectmatrix[12] = 0;

    cl_entities[0].pjectmatrix[1] = 0;
    cl_entities[0].pjectmatrix[5] = (2 * NEARCLIP) / (top - bottom);
    cl_entities[0].pjectmatrix[9] = (top + bottom) / (top - bottom);
    cl_entities[0].pjectmatrix[13] = 0;

    cl_entities[0].pjectmatrix[2] = 0;
    cl_entities[0].pjectmatrix[6] = 0;
    cl_entities[0].pjectmatrix[10] = -1 * nudge;
    cl_entities[0].pjectmatrix[14] = -2 * NEARCLIP * nudge;

    cl_entities[0].pjectmatrix[3] = 0;
    cl_entities[0].pjectmatrix[7] = 0;
    cl_entities[0].pjectmatrix[11] = -1;
    cl_entities[0].pjectmatrix[15] = 0;

    if (waterthisframe)
    {
        int contents = Mod_PointInLeaf (r_origin, cl.worldmodel)->contents;

        if ((contents == CONTENTS_WATER || contents == CONTENTS_SLIME || contents == CONTENTS_LAVA || cl.inwater) && v_blend[3])
        {
            cl_entities[0].pjectmatrix[4] = sin (cl.time) * 0.075f;
            cl_entities[0].pjectmatrix[1] = cos (cl.time) * 0.075f;
        }
    }
    glLoadMatrixf (cl_entities[0].pjectmatrix);
}

/*
=============
R_SetupGL
=============
*/
void R_SetupGL (void)
{
	// glx and gly will always be 0 and 0
	// glwidth and glheight are the resolution we're rendering at
	glViewport (glx, gly, glwidth, glheight);

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

    GL_SetPerspective (r_refdef.fov_x, r_refdef.fov_y);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	// put Z going up and Y going deep
	glRotatef (-90,  1, 0, 0);
	glRotatef (90,  0, 0, 1);

	// set the view according to where the player actually is in the map!!!
	glRotatef (-r_refdef.viewangles[2], 1, 0, 0);
	glRotatef (-r_refdef.viewangles[0], 0, 1, 0);
	glRotatef (-r_refdef.viewangles[1], 0, 0, 1);

	glTranslatef (-r_refdef.vieworg[0], -r_refdef.vieworg[1], -r_refdef.vieworg[2]);

	// set drawing parms
	if (gl_cull.value)
	{
		glEnable (GL_CULL_FACE);
		glCullFace (GL_FRONT);
	}
	else
	{
		glDisable (GL_CULL_FACE);
	}
	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
	//glEnable (GL_TEXTURE_2D); ?????

	// save out a copy of the modelview matrix (used in shadows).
	glGetFloatv (GL_MODELVIEW_MATRIX, cl_entities[0].mvievmatrix);

	// save out a copy of the projection matrix (used for waterwarp).
	glGetFloatv (GL_PROJECTION_MATRIX, cl_entities[0].pjectmatrix);
}


/*
================
R_RenderScenefloat r_world_matrix[16];

r_refdef must be set before the first call
================
*/
void R_RenderScene (void)
{
	R_SetupFrame ();

	R_SetupGL ();

    R_ExtractFrustum ();

	R_MarkLeaves ();	// done here so we know if we're in water

	R_DrawWorld ();		// adds static entities to the list

	S_ExtraUpdate ();	// don't let sound get messed up if going slow

	R_DrawAliasEntities ();

	R_DrawWaterSurfaces ();

	R_DrawSpriteEntities ();

	R_DrawViewModel ();

	R_DrawParticles ();
}

/*
=============
R_Clear

Got weird if you want it.  Realm runs *faster* when clearing the depth buffer every frame... go figure...
=============
*/
void R_Clear (void)
{
	// always accumulate depth buffer to the clear
	GLbitfield QGL_CLEAR_BITFIELD = (GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	// add the colour buffer if we want to explicitly clear
	if (gl_clear.value)
	{
		QGL_CLEAR_BITFIELD |= GL_COLOR_BUFFER_BIT;
	}

	// clear the stencil buffer to 1
	glClearStencil (1);

	// (this should always happen cos we always clear the stencil and depth buffers)
	glClear (QGL_CLEAR_BITFIELD);

	gldepthmin = 0;
	gldepthmax = 1;
	glDepthFunc (GL_LEQUAL);

	glDepthRange (gldepthmin, gldepthmax);
}


/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
void R_RenderView (void)
{
	if (!r_worldentity.model || !cl.worldmodel) return;
	if (r_norefresh.value) return;

	// draw this on frame 3 cos the first few are skipped over
	if (r_framecount == 3)
	{
		// i like this...
		SCR_CenterPrint (cl.levelname);
	}
	r_framecount++;

	// set client frametime
	// base this on realtime cos cl.time doesn't advance when paused, which can cause some weirdness when effects
	// based on cl_frametime update after being unpaused...
	r_frametime = realtime - r_oldframetime;
	r_oldframetime = realtime;

	// do nothing the first coupla frames to give things a chance to settle
	if (r_framecount < 3) return;

	if (!r_worldentity.model || !cl.worldmodel)
	{
		Sys_Error ("R_RenderView: NULL worldmodel");
	}

	if (gl_finish.value)
	{
		glFinish ();
	}
	R_Clear ();

	R_RenderScene ();

	R_BloomBlend ();

	R_PolyBlend ();
}
