/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


#include "gl_state.h"

// includes for image libs
#include "png.h"

gltexture_t texturecache[MAX_GL_TEXTURES];
void GL_InitTextureSlot (gltexture_t *glt);

// MD5 driver interface
void MD5_Checksum (unsigned char *data, int dataLen, unsigned char *checksum);
char MD5_Compare (unsigned char *digest1, unsigned char *digest2);
void MD5_Copy (unsigned char *srcdigest, unsigned char *dstdigest);

/*
================
GL_MakeTextureObjects

Interface into glGenTextures.  Does a few other things as well, including keeping track of the highest
texture name used so we can validly delete textures at video shutdown.
================
*/
GLuint GL_MaxTexture = 0;

void GL_MakeTextureObjects (GLsizei n, GLuint *textures)
{
    int i;

    // make the texture objects first
    // this is the only place glGenTextures is called in the engine
    glGenTextures (n, textures);

    // now check them all and record if the max texture object name has increased
    for (i = 0; i < n; i++)
    {
        if (textures[i] > GL_MaxTexture)
        {
            GL_MaxTexture = textures[i];
        }
    }
}

void GL_DeleteAllTextures (void)
{
    GLuint i;

    for (i = 0; i < GL_MaxTexture; i++)
    {
        if (glIsTexture (i))
        {
            glDeleteTextures (1, &i);
        }
    }
}


/*
==================
GL_UploadTextureToOpenGL

Here we upload our texture!

This is kept fairly generic and flexible to allow reuse for different types of texture.

It's also a direct entry point for certain types of texture.
==================
*/
int GL_UploadTextureToOpenGL (void *data, int width, int height, unsigned int flags)
{
    GLuint	texnum = 0;
    GLenum	InternalFormat;

    // check has the texture already been generated
    if (!(flags & TEX_NOGEN))
    {
        // create a texture object
        GL_MakeTextureObjects (1, &texnum);

        // bind to the texture object
        glBindTexture (GL_TEXTURE_2D, texnum);
    }

    // hardware mipmapping (OpenGL 1.4 or higher an absolute must)
    glTexParameterf (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, (flags & TEX_MIPMAP) ? GL_TRUE : GL_FALSE);

    if (flags & TEX_LIGHTMAP)
        InternalFormat = GL_RGBA;
    else if ((flags & TEX_HASALPHA) && (flags & TEX_NOCOMPRESS))
        InternalFormat = GL_RGBA;
    else if (flags & TEX_HASALPHA)
        InternalFormat = GL_COMPRESSED_RGBA;
    else if (flags & TEX_NOCOMPRESS)
        InternalFormat = GL_RGB;
    else InternalFormat = GL_COMPRESSED_RGB;

    // create the texture
    glTexImage2D (GL_TEXTURE_2D,
                  0,
                  InternalFormat,
                  width,
                  height,
                  0,
                  GL_RGBA,
                  GL_UNSIGNED_BYTE,
                  data);

    // set the filtering modes
    if (flags & TEX_MIPMAP)
    {
        glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    else
    {
        glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    return texnum;
}


/*
=====================
GL_ResampleTextureLerpLine

Resamples an interpolated texture line (I think... not too sure, got this code from somewhere else)
=====================
*/
static void GL_ResampleTextureLerpLine (unsigned char *in, unsigned char *out, int inwidth, int outwidth)
{
    int j, xi, oldx = 0, f, fstep, endx;

    fstep = (int) (inwidth * 65536.0f / outwidth);
    endx = (inwidth - 1);

    for (j = 0, f = 0; j < outwidth; j++, f += fstep)
    {
        xi = (int) f >> 16;

        if (xi != oldx)
        {
            in += (xi - oldx) * 4;
            oldx = xi;
        }

        if (xi < endx)
        {
            int lerp = f & 0xFFFF;

            *out++ = (unsigned char) ((((in[4] - in[0]) * lerp) >> 16) + in[0]);
            *out++ = (unsigned char) ((((in[5] - in[1]) * lerp) >> 16) + in[1]);
            *out++ = (unsigned char) ((((in[6] - in[2]) * lerp) >> 16) + in[2]);
            *out++ = (unsigned char) ((((in[7] - in[3]) * lerp) >> 16) + in[3]);
        }
        else
        {
            // last pixel of the line has no pixel to lerp to
            *out++ = in[0];
            *out++ = in[1];
            *out++ = in[2];
            *out++ = in[3];
        }
    }
}


/*
=================
GL_ResampleTexture

Resize a texture to powers of 2
=================
*/
void GL_ResampleTexture666 (unsigned *indata, int inwidth, int inheight, unsigned *outdata, int outwidth, int outheight)
{
    int				i, j, yi, oldy, f, fstep, endy = (inheight - 1);
    unsigned char	*inrow, *out, *row1, *row2, *memalloc;

    out = (unsigned char *) outdata;
    fstep = (int) (inheight * 65536.0f / outheight);

    memalloc = malloc (2 * (outwidth * 4));
    memset (memalloc, 0, (2 * (outwidth * 4)));
    row1 = memalloc;
    row2 = memalloc + (outwidth * 4);
    inrow = (unsigned char *) indata;
    oldy = 0;

    GL_ResampleTextureLerpLine (inrow, row1, inwidth, outwidth);
    GL_ResampleTextureLerpLine (inrow + inwidth * 4, row2, inwidth, outwidth);

    for (i = 0, f = 0; i < outheight; i++, f += fstep)
    {
        yi = f >> 16;

        if (yi < endy)
        {
            int lerp = f & 0xFFFF;

            if (yi != oldy)
            {
                inrow = (unsigned char *)indata + inwidth * 4 * yi;

                if (yi == oldy + 1)
                    memcpy (row1, row2, outwidth * 4);
                else
                    GL_ResampleTextureLerpLine (inrow, row1, inwidth, outwidth);

                GL_ResampleTextureLerpLine (inrow + inwidth * 4, row2, inwidth, outwidth);

                oldy = yi;
            }

            for (j = outwidth; j; j--)
            {
                out[0] = (unsigned char) ((((row2[0] - row1[0]) * lerp) >> 16) + row1[0]);
                out[1] = (unsigned char) ((((row2[1] - row1[1]) * lerp) >> 16) + row1[1]);
                out[2] = (unsigned char) ((((row2[2] - row1[2]) * lerp) >> 16) + row1[2]);
                out[3] = (unsigned char) ((((row2[3] - row1[3]) * lerp) >> 16) + row1[3]);

                out += 4;
                row1 += 4;
                row2 += 4;
            }
            row1 -= outwidth * 4;
            row2 -= outwidth * 4;
        }
        else
        {
            if (yi != oldy)
            {
                inrow = (unsigned char *) indata + inwidth * 4 * yi;

                if (yi == oldy + 1)
                    memcpy (row1, row2, outwidth * 4);
                else
                    GL_ResampleTextureLerpLine (inrow, row1, inwidth, outwidth);

                oldy = yi;
            }
            memcpy (out, row1, outwidth * 4);
        }
    }
    free (memalloc);
}

typedef struct resamplepoint_s
{
    int p1;
    int p2;
    int p1percent;
    int p2percent;
} resamplepoint_t;

void MakeResamplePoint (resamplepoint_t *resamp, int inscale, int outscale)
{
    int i;

    // set up the resample points across
    for (i = 0; i < outscale; i++)
    {
        // calculate the equalivalent point in the "in" buffer
        float fp = (float) (inscale - 1) * ((float) i / (float) (outscale - 1));

        resamp[i].p1 = (int) floor (fp);
        resamp[i].p2 = (int) ceil (fp);

        // the decimal part is the percentage of p2 used
        resamp[i].p2percent = (int) ((fp - (float) resamp[i].p1) * 100.0f);

        // the rest is the percentage of p2 used
        resamp[i].p1percent = 100 - resamp[i].p2percent;
    }
}


void GL_ResampleTexture (unsigned *in, int inwidth, int inheight, unsigned *out, int outwidth, int outheight)
{
    int		h;
    int		w;
    int		outpoint;
    byte	*indata;
    byte	*outdata;
    int		inrgba[4];
    int		inpercents[4];

    resamplepoint_t *resamplew = NULL;
    resamplepoint_t *resampleh = NULL;

    resamplew = (resamplepoint_t *) malloc (outwidth * sizeof (resamplepoint_t));
    resampleh = (resamplepoint_t *) malloc (outheight * sizeof (resamplepoint_t));

    // set up our resample points in 2 lookup tables for added speed
    MakeResamplePoint (resamplew, inwidth, outwidth);
    MakeResamplePoint (resampleh, inheight, outheight);

    for (h = 0, outpoint = 0; h < outheight; h++)
    {
        for (w = 0; w < outwidth; w++, outpoint++)
        {
            // get a valid pointer to the outdata
            outdata = (byte *) &out[outpoint];

            // set up the percents
            inpercents[0] = (resamplew[w].p1percent + resampleh[h].p1percent);
            inpercents[1] = (resamplew[w].p2percent + resampleh[h].p1percent);
            inpercents[2] = (resamplew[w].p1percent + resampleh[h].p2percent);
            inpercents[3] = (resamplew[w].p2percent + resampleh[h].p2percent);

            // initially 0
            inrgba[0] = inrgba[1] = inrgba[2] = inrgba[3] = 0;

            // accumulate the data for each point
            indata = (byte *) &in[resamplew[w].p1 + resampleh[h].p1 * inwidth];
            inrgba[0] += (int) indata[0] * inpercents[0];
            inrgba[1] += (int) indata[1] * inpercents[0];
            inrgba[2] += (int) indata[2] * inpercents[0];
            inrgba[3] += (int) indata[3] * inpercents[0];

            indata = (byte *) &in[resamplew[w].p2 + resampleh[h].p1 * inwidth];
            inrgba[0] += (int) indata[0] * inpercents[1];
            inrgba[1] += (int) indata[1] * inpercents[1];
            inrgba[2] += (int) indata[2] * inpercents[1];
            inrgba[3] += (int) indata[3] * inpercents[1];

            indata = (byte *) &in[resamplew[w].p1 + resampleh[h].p2 * inwidth];
            inrgba[0] += (int) indata[0] * inpercents[2];
            inrgba[1] += (int) indata[1] * inpercents[2];
            inrgba[2] += (int) indata[2] * inpercents[2];
            inrgba[3] += (int) indata[3] * inpercents[2];

            indata = (byte *) &in[resamplew[w].p2 + resampleh[h].p2 * inwidth];
            inrgba[0] += (int) indata[0] * inpercents[3];
            inrgba[1] += (int) indata[1] * inpercents[3];
            inrgba[2] += (int) indata[2] * inpercents[3];
            inrgba[3] += (int) indata[3] * inpercents[3];

            // save out
            outdata[0] = (byte) (inrgba[0] / 400);
            outdata[1] = (byte) (inrgba[1] / 400);
            outdata[2] = (byte) (inrgba[2] / 400);
            outdata[3] = (byte) (inrgba[3] / 400);
        }
    }
    free (resamplew);
    free (resampleh);
}

/*
==================
GL_Convert8BPPTo32BPP

upsamples the bit depth from 8bpp to 32bpp using the quake palette.
also does some required housekeeping for luma textures.
==================
*/
static void GL_Convert8BPPTo32BPP (gltexture_t *glt, unsigned int *DataOut, qboolean LumaPass)
{
    int		i;
    byte	pix;

    for (i = 0; i < glt->width * glt->height; i++)
    {
        // get the current pixel seperately so any changes we make to it
        // don't screw with our original data
        pix = glt->data[i];

        // set the correct alpha colour
        if (LumaPass && pix < 224)
        {
            // set alpha colour
            DataOut[i] = 0;
        }
        else if (pix == 255)
        {
            // set alpha colour
            DataOut[i] = 0;
        }
        else if (!LumaPass && pix > 223)
        {
            if ((glt->flags & TEX_LIQUID) || !(glt->flags & TEX_MIPMAP))
            {
                // retain the original colour for liquid or non-mipped texes
                DataOut[i] = d_8to24table[pix];
            }
            else
            {
                // set alpha colour
                DataOut[i] = 0;

                // record that it has a luma
                glt->flags |= TEX_HASLUMA;
            }
        }
        else
        {
            DataOut[i] = d_8to24table[pix];
        }
    }
}

/*
===================
GL_SetLiquidColour

Stores an RGB (really RGBA but A is constant) in the glt struct so that the underwater
fog colour can be set correctly to the same colour as the liquid texture.
===================
*/
void GL_SetLiquidColour (gltexture_t *glt, byte *data)
{
    int				i;
    unsigned int	rgba[3] = {0, 0, 0};

    for (i = 0; i < glt->width * glt->height; i++, data += 4)
    {
        rgba[0] += (unsigned int) data[0];
        rgba[1] += (unsigned int) data[1];
        rgba[2] += (unsigned int) data[2];
    }
    glt->rgba[0] = (GLfloat) rgba[0] / (GLfloat) (glt->width * glt->height) / 128.0;
    glt->rgba[1] = (GLfloat) rgba[1] / (GLfloat) (glt->width * glt->height) / 128.0;
    glt->rgba[2] = (GLfloat) rgba[2] / (GLfloat) (glt->width * glt->height) / 128.0;

    // we need to store alpha too so we can pass this into glFogfv
    glt->rgba[3] = 0.7f;
}


/*
===================
GL_AlphaMaskTexture

Converts an RGBA texture to all white with an alpha channel for use by particles, crosshairs, etc
===================
*/
void GL_AlphaMaskTexture (gltexture_t *glt, byte *data)
{
    int i;

    for (i = 0; i < glt->width * glt->height; i++, data += 4)
    {
        // set alpha channel to the value of the red channel
        // (it's assumed these are greyscale)
        data[3] = data[0];

        // set rgb to white
        data[0] = 255;
        data[1] = 255;
        data[2] = 255;
    }
}


/*
===============
GL_QuantumLumaHackery

one popular texture set uses 4/2/4 for it's luma background!!!
while it's really the responsibility of the texture artist to sort this mess out,
it's a pain the face for anyone to do so.  here we just replace them with 0/0/0
===============
*/
void GL_QuantumLumaHackery (gltexture_t *glt, byte *data)
{
    int i;

    for (i = 0; i < glt->width * glt->height; i++, data += 4)
    {
        // use < 10 to catch variations, set to all 0
        if (data[0] < 10 && data[1] < 10 && data[2] < 10)
        {
            data[0] = data[1] = data[2] = 0;
        }
    }
}

/*
=========================================================================================
								PNG IMAGE HANDLING
=========================================================================================
*/

static void PNG_IO_user_read_data (png_structp png, png_bytep data, png_size_t length)
{
    filehandle_t handle = (filehandle_t)png_get_io_ptr(png);
    COM_FRead (data, (int) length, 1, handle);
}

/*
===============
GL_LoadPNG

PNG Loading
===============
*/
byte *GL_LoadPNG (gltexture_t *glt, char *basename)
{
    unsigned char	header[8], **rowpointers;
    png_structp		png;
    png_infop		pnginfo;
    int				y, width, height, bitdepth, colortype;
    int				interlace, compression, filter, bytesperpixel;
    unsigned long	rowbytes;
    byte			*ImageRGBAData;
    filehandle_t	ImgFile;
    char			loadname[256];

    // va sucks dogs cocks in hell
    va_snprintf ("GL_LoadPNG", loadname, sizeof(loadname), "%s.png", basename);

    COM_FOpenFile (loadname, &ImgFile, FS_READ);

    if (!ImgFile)
    {
        return NULL;
    }
    COM_FRead (header, 8, 1, ImgFile);

    if (png_sig_cmp (header, 0, 8) != 0)
    {
        Con_SafePrintf ("%s : Invalid PNG file\n", basename);
        goto BadPNG;
    }

    if (!(png = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)))
    {
        goto BadPNG;
    }

    if (!(pnginfo = png_create_info_struct (png)))
    {
        png_destroy_read_struct (&png, &pnginfo, NULL);
        goto BadPNG;
    }

    if (setjmp (png->jmpbuf))
    {
        png_destroy_read_struct (&png, &pnginfo, NULL);
        goto BadPNG;
    }
    png_set_read_fn (png, (png_voidp)ImgFile, PNG_IO_user_read_data);
    png_set_sig_bytes (png, 8);
    png_read_info (png, pnginfo);
    png_get_IHDR (png, pnginfo, (png_uint_32 *) &width, (png_uint_32 *) &height, &bitdepth, &colortype, &interlace, &compression, &filter);

    if (colortype == PNG_COLOR_TYPE_PALETTE)
    {
        png_set_palette_to_rgb (png);
        png_set_filler (png, 255, PNG_FILLER_AFTER);
    }

    if (colortype == PNG_COLOR_TYPE_GRAY && bitdepth < 8)
    {
        png_set_expand_gray_1_2_4_to_8 (png);
    }

    if (png_get_valid(png, pnginfo, PNG_INFO_tRNS))
    {
        png_set_tRNS_to_alpha (png);
    }

    if (colortype == PNG_COLOR_TYPE_GRAY || colortype == PNG_COLOR_TYPE_GRAY_ALPHA)
    {
        png_set_gray_to_rgb (png);
        png_set_filler (png, 255, PNG_FILLER_AFTER);
    }

    if (colortype != PNG_COLOR_TYPE_RGBA)
    {
        png_set_filler (png, 255, PNG_FILLER_AFTER);
    }

    if (bitdepth < 8)
    {
        png_set_expand (png);
    }
    else if (bitdepth == 16)
    {
        png_set_strip_16 (png);
    }
    png_read_update_info (png, pnginfo);
    rowbytes = png_get_rowbytes (png, pnginfo);
    bytesperpixel = png_get_channels (png, pnginfo);
    bitdepth = png_get_bit_depth (png, pnginfo);

    // fill in properties
    // just using width and height are *not* good here...!
    // rowbytes / 4 should also be good for use here, but we'll do it the official way
    glt->width = png_get_image_width (png, pnginfo);
    glt->height = png_get_image_height (png, pnginfo);

    if (bitdepth != 8 || bytesperpixel != 4)
    {
        png_destroy_read_struct (&png, &pnginfo, NULL);
        goto BadPNG;
    }
    ImageRGBAData = malloc (height * rowbytes);
    rowpointers = malloc (height * 4);

    for (y = 0; y < height; y++)
    {
        rowpointers[y] = ImageRGBAData + y * rowbytes;
    }
    png_read_image (png, rowpointers);
    png_read_end (png, NULL);

    png_destroy_read_struct (&png, &pnginfo, NULL);
    free (rowpointers);
    COM_FCloseFile (ImgFile);
    return ImageRGBAData;

BadPNG:
    ;
    COM_FCloseFile (ImgFile);
    return NULL;
}

/*
=========================================================================================
								TGA IMAGE HANDLING
=========================================================================================
*/

#define TGA_MAXCOLORS 16384

/* Definitions for image types. */
#define TGA_Null	0	/* no image data */
#define TGA_Map		1	/* Uncompressed, color-mapped images. */
#define TGA_RGB		2	/* Uncompressed, RGB images. */
#define TGA_Mono	3	/* Uncompressed, black and white images. */
#define TGA_RLEMap	9	/* Runlength encoded color-mapped images. */
#define TGA_RLERGB	10	/* Runlength encoded RGB images. */
#define TGA_RLEMono	11	/* Compressed, black and white images. */
#define TGA_CompMap	32	/* Compressed color-mapped data, using Huffman, Delta, and runlength encoding. */
#define TGA_CompMap4	33	/* Compressed color-mapped data, using Huffman, Delta, and runlength encoding. 4-pass quadtree-type process. */

/* Definitions for interleave flag. */
#define TGA_IL_None	0	/* non-interleaved. */
#define TGA_IL_Two	1	/* two-way (even/odd) interleaving */
#define TGA_IL_Four	2	/* four way interleaving */
#define TGA_IL_Reserved	3	/* reserved */

/* Definitions for origin flag */
#define TGA_O_UPPER	0	/* Origin in lower left-hand corner. */
#define TGA_O_LOWER	1	/* Origin in upper left-hand corner. */

typedef struct TGAHeader_s
{
    unsigned char id_length;
    unsigned char colormap_type;
    unsigned char image_type;
    unsigned short colormap_index;
    unsigned short colormap_length;
    unsigned char colormap_size;
    unsigned short x_origin;
    unsigned short y_origin;
    unsigned short width;
    unsigned short height;
    unsigned char pixel_size;
    unsigned char attributes;
} TGAHeader_t;

/*
=============
GL_LoadTGA

TGA Loading stuff
=============
*/
byte *GL_LoadTGA (gltexture_t *glt, char *basename)
{
    int				w, h, x, y, realrow, truerow, baserow, i, temp1, temp2, pixel_size, map_idx;
    int				RLE_count, RLE_flag, size, interleave, origin;
    qboolean		mapped, rlencoded;
    byte			*ImageRGBAData, *dst, r, g, b, a, j, k, l, *ColorMap;
    TGAHeader_t		header;
    filehandle_t	ImgFile;
    char			loadname[256];

    // va sucks dogs cocks in hell
    va_snprintf ("GL_LoadTGA", loadname, sizeof(loadname), "%s.tga", basename);

    COM_FOpenFile (loadname, &ImgFile, FS_READ);

    if (!ImgFile)
    {
        return NULL;
    }
    header.id_length = COM_FGetc (ImgFile);
    header.colormap_type = COM_FGetc (ImgFile);
    header.image_type = COM_FGetc (ImgFile);
    header.colormap_index = COM_FGetcLittleShort (ImgFile);
    header.colormap_length = COM_FGetcLittleShort (ImgFile);
    header.colormap_size = COM_FGetc (ImgFile);
    header.x_origin = COM_FGetcLittleShort (ImgFile);
    header.y_origin = COM_FGetcLittleShort (ImgFile);
    header.width = COM_FGetcLittleShort (ImgFile);
    header.height = COM_FGetcLittleShort (ImgFile);
    header.pixel_size = COM_FGetc (ImgFile);
    header.attributes = COM_FGetc (ImgFile);

    if (header.id_length != 0)
    {
        COM_FSeek (ImgFile, header.id_length, FS_SEEK_CUR);
    }

    /* validate TGA type */
    switch (header.image_type)
    {
    case TGA_Map:
    case TGA_RGB:
    case TGA_Mono:
    case TGA_RLEMap:
    case TGA_RLERGB:
    case TGA_RLEMono:
        break;

    default:
        Con_SafePrintf ("Unsupported TGA image %s: Only type 1 (map), 2 (RGB), 3 (mono), 9 (RLEmap), 10 (RLERGB), 11 (RLEmono) TGA images supported\n", COM_SkipPath(basename));
        goto BadTGA;
    }

    /* validate color depth */
    switch (header.pixel_size)
    {
    case 8:
    case 15:
    case 16:
    case 24:
    case 32:
        break;

    default:
        Con_SafePrintf ("Unsupported TGA image %s: Only 8, 15, 16, 24 or 32 bit images (with colormaps) supported\n", COM_SkipPath(basename));
        goto BadTGA;
    }
    r = g = b = a = l = 0;

    /* if required, read the color map information. */
    ColorMap = NULL;
    mapped = (header.image_type == TGA_Map || header.image_type == TGA_RLEMap) && header.colormap_type == 1;

    if (mapped)
    {
        /* validate colormap size */
        switch (header.colormap_size)
        {
        case 8:
        case 15:
        case 16:
        case 32:
        case 24:
            break;

        default:
            Con_SafePrintf ("Unsupported TGA image %s: Only 8, 15, 16, 24 or 32 bit colormaps supported\n", COM_SkipPath(basename));
            goto BadTGA;
        }
        temp1 = header.colormap_index;
        temp2 = header.colormap_length;

        if ((temp1 + temp2 + 1) >= TGA_MAXCOLORS)
        {
            goto BadTGA;
        }
        ColorMap = malloc (TGA_MAXCOLORS * 4);
        map_idx = 0;

        for (i = temp1 ; i < temp1 + temp2 ; ++i, map_idx += 4)
        {
            /* read appropriate number of bytes, break into rgba & put in map. */
            switch (header.colormap_size)
            {
            case 8:	/* grey scale, read and triplicate. */
                r = g = b = COM_FGetc (ImgFile);
                a = 255;
                break;

            case 15:	/* 5 bits each of red green and blue. */
                /* watch byte order. */
                j = COM_FGetc (ImgFile);
                k = COM_FGetc (ImgFile);
                l = ((unsigned int)k << 8) + j;
                r = (byte)(((k & 0x7C) >> 2) << 3);
                g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
                b = (byte)((j & 0x1F) << 3);
                a = 255;
                break;

            case 16:	/* 5 bits each of red green and blue, 1 alpha bit. */
                /* watch byte order. */
                j = COM_FGetc (ImgFile);
                k = COM_FGetc (ImgFile);
                l = ((unsigned int)k << 8) + j;
                r = (byte)(((k & 0x7C) >> 2) << 3);
                g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
                b = (byte)((j & 0x1F) << 3);
                a = (k & 0x80) ? 255 : 0;
                break;

            case 24:	/* 8 bits each of blue, green and red. */
                b = COM_FGetc (ImgFile);
                g = COM_FGetc (ImgFile);
                r = COM_FGetc (ImgFile);
                a = 255;
                l = 0;
                break;

            case 32:	/* 8 bits each of blue, green, red and alpha. */
                b = COM_FGetc (ImgFile);
                g = COM_FGetc (ImgFile);
                r = COM_FGetc (ImgFile);
                a = COM_FGetc (ImgFile);
                l = 0;
                break;
            }
            ColorMap[map_idx+0] = r;
            ColorMap[map_idx+1] = g;
            ColorMap[map_idx+2] = b;
            ColorMap[map_idx+3] = a;
        }
    }

    /* check run-length encoding. */
    rlencoded = (header.image_type == TGA_RLEMap || header.image_type == TGA_RLERGB || header.image_type == TGA_RLEMono);
    RLE_count = RLE_flag = 0;

    // fill in properties
    glt->width = w = header.width;
    glt->height = h = header.height;

    size = w * h * 4;
    ImageRGBAData = calloc (size, 1);

    /* read the Targa file body and convert to portable format. */
    pixel_size = header.pixel_size;
    origin = (header.attributes & 0x20) >> 5;
    interleave = (header.attributes & 0xC0) >> 6;
    truerow = baserow = 0;

    for (y=0 ; y<h ; y++)
    {
        realrow = truerow;

        if (origin == TGA_O_UPPER)
        {
            realrow = h - realrow - 1;
        }
        dst = ImageRGBAData + realrow * w * 4;

        for (x=0 ; x<w ; x++)
        {
            /* check if run length encoded. */
            if (rlencoded)
            {
                if (!RLE_count)
                {
                    /* have to restart run. */
                    i = COM_FGetc (ImgFile);

                    RLE_flag = (i & 0x80);

                    if (!RLE_flag)	// stream of unencoded pixels
                    {
                        RLE_count = i + 1;
                    }
                    else		// single pixel replicated
                    {
                        RLE_count = i - 127;
                    }

                    /* decrement count & get pixel. */
                    --RLE_count;
                }
                else
                {
                    /* have already read count & (at least) first pixel. */
                    --RLE_count;

                    if (RLE_flag)
                    {
                        /* replicated pixels. */
                        goto PixEncode;
                    }
                }
            }

            /* read appropriate number of bytes, break into RGB. */
            switch (pixel_size)
            {
            case 8:	/* grey scale, read and triplicate. */
                r = g = b = l = COM_FGetc (ImgFile);
                a = 255;
                break;

            case 15:	/* 5 bits each of red green and blue. */
                /* watch byte order. */
                j = COM_FGetc (ImgFile);
                k = COM_FGetc (ImgFile);
                l = ((unsigned int)k << 8) + j;
                r = (byte)(((k & 0x7C) >> 2) << 3);
                g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
                b = (byte)((j & 0x1F) << 3);
                a = 255;
                break;

            case 16:	/* 5 bits each of red green and blue, 1 alpha bit. */
                /* watch byte order. */
                j = COM_FGetc (ImgFile);
                k = COM_FGetc (ImgFile);
                l = ((unsigned int)k << 8) + j;
                r = (byte)(((k & 0x7C) >> 2) << 3);
                g = (byte)((((k & 0x03) << 3) + ((j & 0xE0) >> 5)) << 3);
                b = (byte)((j & 0x1F) << 3);
                a = (k & 0x80) ? 255 : 0;
                break;

            case 24:	/* 8 bits each of blue, green and red. */
                b = COM_FGetc (ImgFile);
                g = COM_FGetc (ImgFile);
                r = COM_FGetc (ImgFile);
                a = 255;
                l = 0;
                break;

            case 32:	/* 8 bits each of blue, green, red and alpha. */
                b = COM_FGetc (ImgFile);
                g = COM_FGetc (ImgFile);
                r = COM_FGetc (ImgFile);
                a = COM_FGetc (ImgFile);
                l = 0;
                break;

            default:
                Con_SafePrintf ("Malformed TGA image %s: Illegal pixel_size '%d'\n", COM_SkipPath(basename), pixel_size);
                COM_FCloseFile (ImgFile);
                free (ImageRGBAData);

                if (mapped)
                {
                    free (ColorMap);
                }
                return NULL;
            }

PixEncode:
            if (mapped)
            {
                map_idx = l * 4;
                *dst++ = ColorMap[map_idx+0];
                *dst++ = ColorMap[map_idx+1];
                *dst++ = ColorMap[map_idx+2];
                *dst++ = ColorMap[map_idx+3];
            }
            else
            {
                *dst++ = r;
                *dst++ = g;
                *dst++ = b;
                *dst++ = a;
            }
        }

        if (interleave == TGA_IL_Four)
        {
            truerow += 4;
        }
        else if (interleave == TGA_IL_Two)
        {
            truerow += 2;
        }
        else
        {
            truerow++;
        }

        if (truerow >= h)
        {
            truerow = ++baserow;
        }
    }

    if (mapped)
    {
        free (ColorMap);
    }
    COM_FCloseFile (ImgFile);
    return ImageRGBAData;

BadTGA:
    ;
    Con_SafePrintf ("Bad TGA: %s\n", loadname);
    COM_FCloseFile (ImgFile);
    return NULL;
}

/*
=================
GL_LoadBMP

BMP Loading Malarkey.  God, this format is horrible...!
=================
*/
#ifndef BI_RGB
#define BI_RGB 0L
#endif

#ifndef BI_RLE8
#define BI_RLE8 1L
#endif

// MH - preferred to ID's version in the q3a source as mine supports compression and 4bpp
byte *GL_LoadBMP (gltexture_t *glt, char *basename)
{
    unsigned char		*ImageRGBAData;
    int					i, j;
    int					run;
    int					runlength;
    unsigned char		bmIn;
    BITMAPINFO			bmInfo;
    BITMAPFILEHEADER	bmHeader;
    RGBQUAD				bmPalette[256];
    unsigned char		*bmData;
    int					pad;
    filehandle_t		ImgFile;
    char				loadname[256];

    // va sucks dogs cocks in hell
    va_snprintf ("GL_LoadBMP", loadname, sizeof(loadname), "%s.bmp", basename);

    COM_FOpenFile (loadname, &ImgFile, FS_READ);

    if (!ImgFile)
    {
        return NULL;
    }

    // read the file header
    COM_FRead (&bmHeader, sizeof (bmHeader), 1, ImgFile);

    // validate the type
    if (bmHeader.bfType != 0x4D42)
    {
        goto BadBMP;
    }

    // read the BITMAPINFO
    COM_FRead (&bmInfo, sizeof (bmInfo), 1, ImgFile);

    if (bmInfo.bmiHeader.biCompression != BI_RGB)
    {
        if (!(bmInfo.bmiHeader.biCompression == BI_RLE8 && bmInfo.bmiHeader.biBitCount == 8))
        {
            goto BadBMP;
        }
    }

    // support 4, 8 and 24 BPP modes
    if (bmInfo.bmiHeader.biBitCount != 4 && bmInfo.bmiHeader.biBitCount != 8 && bmInfo.bmiHeader.biBitCount != 24)
    {
        Con_SafePrintf ("Unsupported BMP: %s\n", loadname);
        goto BadBMP;
    }
    ImageRGBAData = malloc (bmInfo.bmiHeader.biWidth * abs (bmInfo.bmiHeader.biHeight) * 4);
    bmData = ImageRGBAData;

    if (bmInfo.bmiHeader.biBitCount != 24)
    {
        // read the palette entries
        // the first one is in the BITMAPINFO struct
        bmPalette[0].rgbBlue = bmInfo.bmiColors[0].rgbBlue;
        bmPalette[0].rgbGreen = bmInfo.bmiColors[0].rgbGreen;
        bmPalette[0].rgbRed = bmInfo.bmiColors[0].rgbRed;
        bmPalette[0].rgbReserved = bmInfo.bmiColors[0].rgbReserved;

        // remaining entries follow the struct
        for (i = 1; i < (int)bmInfo.bmiHeader.biClrUsed; i++)
        {
            COM_FRead (&bmPalette[i], sizeof (RGBQUAD), 1, ImgFile);
        }
    }

    // each row in a bitmap is padded to align it on a 4 unsigned char boundary, so find
    // out how much padding we need (this may not work with RLE images)
    for (i = 0; ; i++)
    {
        pad = i * 4;
        if (pad >= (bmInfo.bmiHeader.biWidth * 3)) break;
    }
    pad -= bmInfo.bmiHeader.biWidth * 3;

    // now seek to the proper offset of the data
    COM_FSeek (ImgFile, bmHeader.bfOffBits, FS_SEEK_SET);

    // load the data
    for (i = 0; i < bmInfo.bmiHeader.biWidth * abs (bmInfo.bmiHeader.biHeight); i++)
    {
        if (bmInfo.bmiHeader.biBitCount == 4)
        {
            bmIn = COM_FGetc (ImgFile);

            bmData[0] = bmPalette[bmIn >> 4].rgbRed;
            bmData[1] = bmPalette[bmIn >> 4].rgbGreen;
            bmData[2] = bmPalette[bmIn >> 4].rgbBlue;
            bmData[3] = 255;

            bmData[4] = bmPalette[bmIn % 16].rgbRed;
            bmData[5] = bmPalette[bmIn % 16].rgbGreen;
            bmData[6] = bmPalette[bmIn % 16].rgbBlue;
            bmData[7] = 255;

            bmData += 8;

            // increment i again cos we're reading two entires at a time
            i++;

            continue;
        }
        else if (bmInfo.bmiHeader.biBitCount == 8)
        {
            if (bmInfo.bmiHeader.biCompression == BI_RLE8)
            {
                bmIn = COM_FGetc (ImgFile);

                if (bmIn == 0)
                {
                    bmIn = COM_FGetc (ImgFile);
                    i++;

                    switch (bmIn)
                    {
                    case 0:
                        // End of Line - just do nothing - trust me, it works.
                        break;
                    case 1:
                        // End of Bitmap
                        goto bmDone;
                        break;
                    case 2:
                        // Delta
                        // (tests indicate this isn't actually used... most likely cos the bitmap format
                        // doesn't specify what the skipped pixels are to be filled with!!!)
                        bmIn = COM_FGetc (ImgFile);
                        i++;

                        bmData += bmIn * 4;

                        bmIn = COM_FGetc (ImgFile);
                        i++;

                        bmData += bmIn * bmInfo.bmiHeader.biWidth * 4;

                        break;
                    default:
                        // absolute mode
                        runlength = bmIn;

                        for (run = 0; run < runlength; run++)
                        {
                            bmIn = COM_FGetc (ImgFile);
                            i++;

                            bmData[0] = bmPalette[bmIn].rgbRed;
                            bmData[1] = bmPalette[bmIn].rgbGreen;
                            bmData[2] = bmPalette[bmIn].rgbBlue;
                            bmData[3] = 255;

                            bmData += 4;
                        }

                        if (runlength % 2)
                        {
                            // align on word boundaries
                            bmIn = COM_FGetc (ImgFile);
                            i++;
                        }

                        break;
                    }
                }
                else
                {
                    // encoded mode
                    runlength = bmIn;

                    bmIn = COM_FGetc (ImgFile);
                    i++;

                    for (run = 0; run < runlength; run++)
                    {
                        bmData[0] = bmPalette[bmIn].rgbRed;
                        bmData[1] = bmPalette[bmIn].rgbGreen;
                        bmData[2] = bmPalette[bmIn].rgbBlue;
                        bmData[3] = 255;

                        bmData += 4;
                    }
                }

                continue;
            }
            bmIn = COM_FGetc (ImgFile);

            bmData[0] = bmPalette[bmIn].rgbRed;
            bmData[1] = bmPalette[bmIn].rgbGreen;
            bmData[2] = bmPalette[bmIn].rgbBlue;
            bmData[3] = 255;

            bmData += 4;

            continue;
        }
        else
        {
            // 24 BPP
            bmData[2] = COM_FGetc (ImgFile);
            bmData[1] = COM_FGetc (ImgFile);
            bmData[0] = COM_FGetc (ImgFile);
            bmData[3] = 255;

            bmData += 4;
        }

        // read the padding
        for (j = 0; j < pad; j++)
        {
            COM_FGetc (ImgFile);
        }
    }

bmDone:
    ;

    // check orientation
    if (bmInfo.bmiHeader.biHeight > 0)
    {
        // bottom-up bitmap so we must invert it (sigh)
        for (i = 0; i < abs (bmInfo.bmiHeader.biHeight) / 2; i++)
        {
            for (j = 0; j < bmInfo.bmiHeader.biWidth; j++)
            {
                unsigned int srcpos = (i * bmInfo.bmiHeader.biWidth) + j;
                unsigned int dstpos = (((abs (bmInfo.bmiHeader.biHeight) - 1) - i) * bmInfo.bmiHeader.biWidth) + j;
                unsigned int tmp;

                tmp = ((unsigned int *) ImageRGBAData)[dstpos];
                ((unsigned int *) ImageRGBAData)[dstpos] = ((unsigned int *) ImageRGBAData)[srcpos];
                ((unsigned int *) ImageRGBAData)[srcpos] = tmp;
            }
        }
    }

    // fill in properties
    glt->width = bmInfo.bmiHeader.biWidth;
    glt->height = abs (bmInfo.bmiHeader.biHeight);
    COM_FCloseFile (ImgFile);
    return ImageRGBAData;

BadBMP:
    ;
    Con_SafePrintf ("Bad BMP: %s\n", loadname);
    COM_FCloseFile (ImgFile);
    return NULL;
}

/*
===============
GL_LoadPCX

PCX Loading
===============
*/
typedef struct pcx_s
{
    char manufacturer;
    char version;
    char encoding;
    char bits_per_pixel;

    unsigned short xmin;
    unsigned short ymin;
    unsigned short xmax;
    unsigned short ymax;

    unsigned short hres;
    unsigned short vres;

    unsigned char palette[48];

    char reserved;
    char color_planes;

    unsigned short bytes_per_line;
    unsigned short palette_type;

    char filler[58];

    unsigned int data;			// unbounded
} pcx_t;


byte *GL_LoadPCX (gltexture_t *glt, char *basename)
{
    pcx_t			*pcx, pcxbuf;
    unsigned char	palette[768];
    unsigned char	*pix;
    int				x, y;
    int				dataByte, runLength;
    int				count;
    byte			*ImageRGBAData;
    filehandle_t	ImgFile;
    char			loadname[256];

    // va sucks dogs cocks in hell
    va_snprintf ("GL_LoadPCX", loadname, sizeof(loadname), "%s.pcx", basename);

    COM_FOpenFile (loadname, &ImgFile, FS_READ);

    if (!ImgFile)
    {
        return NULL;
    }

    // parse the PCX file
    COM_FRead (&pcxbuf, sizeof (pcxbuf), 1, ImgFile);

    pcx = &pcxbuf;

    if (pcx->manufacturer != 0x0a || pcx->version != 5 || pcx->encoding != 1 || pcx->bits_per_pixel != 8 || pcx->xmax >= 1024 || pcx->ymax >= 1024)
    {
        goto BadPCX;
    }

    // seek to palette
    COM_FSeek (ImgFile, -768, FS_SEEK_END);
    COM_FRead (palette, 768, 1, ImgFile);
    COM_FSeek (ImgFile, sizeof (pcxbuf) - 4, FS_SEEK_SET);

    count = (pcx->xmax + 1) * (pcx->ymax + 1);
    ImageRGBAData = malloc (count * 4);

    for (y = 0; y <= pcx->ymax; y++)
    {
        pix = ImageRGBAData + 4 * y * (pcx->xmax + 1);

        for (x = 0; x <= pcx->xmax;)
        {
            dataByte = COM_FGetc (ImgFile);

            if ((dataByte & 0xC0) == 0xC0)
            {
                runLength = dataByte & 0x3F;
                dataByte = COM_FGetc (ImgFile);
            }
            else
            {
                runLength = 1;
            }

            while(runLength-- > 0)
            {
                pix[0] = palette[dataByte * 3];
                pix[1] = palette[dataByte * 3 + 1];
                pix[2] = palette[dataByte * 3 + 2];

                // PCX files never have alpha
                pix[3] = 255;
                pix += 4;
                x++;
            }
        }
    }

    // fill in properties
    glt->width = pcx->xmax + 1;
    glt->height = pcx->ymax + 1;
    COM_FCloseFile (ImgFile);
    return ImageRGBAData;

BadPCX:
    ;
    COM_FCloseFile (ImgFile);
    return NULL;
}

/*
==============
GL_TryForExternalTexture

Attempts to load an external texture replacement.
==============
*/
byte *GL_TryForExternalTexture (gltexture_t *glt, const char *basename)
{
    // fun with function pointers...
    byte	*(*GL_Texloader[4]) (gltexture_t *, char *) = {GL_LoadPNG, GL_LoadTGA, GL_LoadBMP, GL_LoadPCX};
    int		i;
    int		j;
    char	TexName[128];
    byte	*texdata = NULL;

    for (j = 0; j < 2; j++)
    {
        for (i = 0; i < 4; i++)
        {
            // basename goes corrupt in the loader cos of nested va's using the same static buffer, so work around it.
            // not everything is loaded from "textures" (e.g. mapshots and skyboxes) so we must allow for that too.
            // however, we prefer "textures" for the first pass cos they're more likely to be there
            if (j)
            {
                va_snprintf ("GL_TryForExternalTexture", TexName, sizeof(TexName), "%s", basename);
            }
            else
            {
                va_snprintf ("GL_TryForExternalTexture", TexName, sizeof(TexName), "textures/%s", basename);
            }
            texdata = GL_Texloader[i] (glt, TexName);

            if (texdata)
            {
                // mark it as external
                glt->flags |= (TEX_EXTERN | TEX_32BPP);

                // got it
                return texdata;
            }

            // this should never happen
            if (texdata)
            {
                free (texdata);
            }
        }
    }

    // didn't get it
    return NULL;
}

/*
==============================================

	NORMAL MAP GENERATION

==============================================
*/
// pixel struct
typedef struct _pixel
{
    byte red;
    byte green;
    byte blue;
    byte alpha;
} pixel;

typedef union _pixelint
{
    pixel			rgba;
    unsigned int	ui;
} pixelint;

void WritePixel (unsigned int *image, unsigned int *pix, int x, int y, int w)
{
    *(image + w * y + x) = *pix;
}

static float brightest;
static float darkest;

int BoxFilter (unsigned int *image, pixelint *pix, int x, int y, int w)
{
    float gs;

    pix->ui = *(image + w * y + x);

    // write greyscale to alpha - NTSC standard
    // the ATI sample code i based this on only used the red channel, which gave wild results on
    // some texes.  this is much more balanced
    gs = (((int) pix->rgba.red) * 30 + ((int) pix->rgba.green) * 59 + ((int) pix->rgba.blue) * 11) / 100;

    // rescale so that the darkest to brightest range is properly 0 to 255
    gs -= darkest;
    gs *= brightest;

    // bound
    if (gs > 255) gs = 255;
    if (gs < 0) gs = 0;

    // take the square root of it to get a better range
    gs /= 255.0;
    gs = sqrt (gs);
    gs *= 255.0;

    return (int) gs;
}

void ReadPixel (unsigned int *image, pixelint *pix, int x, int y, int w)
{
    int accum = 0;

    // accumulate the 4 pixels
    accum += BoxFilter (image, pix, x, y, w);
    accum += BoxFilter (image, pix, x, y + 1, w);
    accum += BoxFilter (image, pix, x + 1, y, w);
    accum += BoxFilter (image, pix, x + 1, y + 1, w);

    pix->rgba.alpha = (byte) (accum / 4);
}


byte PackFloatInByte (float in)
{
    return (byte) ((in + 1.0f) / 2.0f * 255.0f);
}


void MakeDOT3 (gltexture_t *glt, unsigned int *src, int width, int height)
{
    unsigned int	*dst = NULL;
    int				x;
    int				y;
    pixelint		pix;
    float			dX, dY, nX, nY, nZ, oolen;
    char			normalmapname[256];

    // use sprintf to avoid pointer confusion from multiple nested va's
    va_snprintf ("MakeDOT3", normalmapname, sizeof(normalmapname), "%s_norm", glt->identifier);

    // try for an external normal map
    dst = (unsigned int *) GL_TryForExternalTexture (glt, normalmapname);

    // we got one, so upload it
    if (dst)
    {
        int w, h;

        // get power of 2 dimensions
        for (w = 1; w < glt->width; w *= 2);
        for (h = 1; h < glt->height; h *= 2);

        // JESUS!  I never tested this in the old one...
        if (w > GLCaps.GL_MaxTextureSize) w = GLCaps.GL_MaxTextureSize;
        if (h > GLCaps.GL_MaxTextureSize) h = GLCaps.GL_MaxTextureSize;

        // now check for resizing
        if (w != glt->width || h != glt->height)
        {
            // allocated a buffer to hold the resampled data
            unsigned int *scaled = (unsigned int *) malloc (w * h * sizeof (unsigned int));

            // resize the texture
            GL_ResampleTexture (dst, glt->width, glt->height, scaled, w, h);

            // switch the data pointer
            free (dst);
            dst = scaled;
        }

        // now send it up to OpenGL
        glt->normtexnum = GL_UploadTextureToOpenGL (dst, w, h, (TEX_MIPMAP | TEX_NOCOMPRESS));

        free (dst);
        return;
    }

    // no external normal map so generate one from the texture data
    dst = (unsigned int *) malloc (width * height * sizeof (unsigned int) / 4);

    for (y = 0; y < height; y += 2)
    {
        for (x = 0; x < width; x += 2)
        {
            // Do Y Sobel filter
            ReadPixel (src, &pix, (x - 2 + width) % width, (y + 2) % height, width);
            dY  = ((float) pix.rgba.alpha) / 255.0f * -1.0f;

            ReadPixel (src, &pix, x % width, (y + 2) % height, width);
            dY += ((float) pix.rgba.alpha) / 255.0f * -2.0f;

            ReadPixel (src, &pix, (x + 2) % width, (y + 2) % height, width);
            dY += ((float) pix.rgba.alpha) / 255.0f * -1.0f;

            ReadPixel (src, &pix, (x - 2 + width) % width, (y - 2 + height) % height, width);
            dY += ((float) pix.rgba.alpha) / 255.0f *  1.0f;

            ReadPixel (src, &pix, x % width, (y - 2 + height) % height, width);
            dY += ((float) pix.rgba.alpha) / 255.0f *  2.0f;

            ReadPixel (src, &pix, (x + 2) % width, (y - 2 + height) % height, width);
            dY += ((float) pix.rgba.alpha) / 255.0f *  1.0f;

            // Do X Sobel filter
            ReadPixel (src, &pix, (x - 2 + width) % width, (y - 2 + height) % height, width);
            dX  = ((float) pix.rgba.alpha) / 255.0f * -1.0f;

            ReadPixel (src, &pix, (x - 2 + width) % width, y % height, width);
            dX += ((float) pix.rgba.alpha) / 255.0f * -2.0f;

            ReadPixel (src, &pix, (x - 2 + width) % width, (y + 2) % height, width);
            dX += ((float) pix.rgba.alpha) / 255.0f * -1.0f;

            ReadPixel (src, &pix, (x + 2) % width, (y - 2 + height) % height, width);
            dX += ((float) pix.rgba.alpha) / 255.0f *  1.0f;

            ReadPixel (src, &pix, (x + 2) % width, y % height, width);
            dX += ((float) pix.rgba.alpha) / 255.0f *  2.0f;

            ReadPixel (src, &pix, (x + 2) % width, (y + 2) % height, width);
            dX += ((float) pix.rgba.alpha) / 255.0f *  1.0f;

            // Cross Product of components of gradient reduces to
            // this is a suitable normal map for opengl - the direct3d equivalent (which ati assumes) is commented out
            nX = dY; //-dX;
            nY = -dX; //-dY;
            nZ = 1;

            // Normalize - nZ is always 1
            oolen = 1.0f / ((float) sqrt (nX * nX + nY * nY + nZ * nZ));
            nX *= oolen;
            nY *= oolen;
            nZ *= oolen;

            pix.rgba.red   = (byte) PackFloatInByte (nX);
            pix.rgba.green = (byte) PackFloatInByte (nY);
            pix.rgba.blue  = (byte) PackFloatInByte (nZ);

            WritePixel (dst, &pix.ui, x / 2, y / 2, width / 2);
        }
    }

    // upload the normal map data with specific properties
    glt->normtexnum = GL_UploadTextureToOpenGL (dst, width / 2, height / 2, (TEX_MIPMAP | TEX_NOCOMPRESS));

    // free dst
    free (dst);
}


// gross evil and vile hack to conjure up a doom 3/quake 4 style specular map from the original texture.
// convert to greyscale, divide by 16 and square the result.
void MakeSpec (gltexture_t *glt, unsigned int *src, int width, int height)
{
    byte	*pix;
    int		i;
    float	specgrey;
    float	origgrey;
    int		specexp = 2;
    int		j;

    for (i = 0; i < height * width; i++)
    {
        pix = (byte *) &src[i];

        // greyscale ntsc standard
        origgrey = ((int) pix[0] * 30 + (int) pix[1] * 59 + (int) pix[2] * 11) / 100;

        // rescale to 0 to 1 scale
        specgrey = origgrey / 255.0f;

        // square it
        for (j = 0; j < specexp; j++)
        {
            specgrey *= specgrey;
        }

        // rescale to 0 to 255 scale
        specgrey *= 255.0f;

        // paranoid bounds checking
        if (specgrey > 255) specgrey = 255;
        if (specgrey < 0) specgrey = 0;

        // now add it to the original texture
        for (j = 0; j < 3; j++)
        {
            if ((int) pix[j] + specgrey > 255)
            {
                pix[j] = 255;
            }
            else
            {
                pix[j] += specgrey;
            }
        }

        // recalc the greyscale for the new pixels
        specgrey = ((int) pix[0] * 30 + (int) pix[1] * 59 + (int) pix[2] * 11) / 100;

        // and bring back to the same intensity as the original
        pix[0] = (byte) ((float) pix[0] * origgrey / specgrey);
        pix[1] = (byte) ((float) pix[1] * origgrey / specgrey);
        pix[2] = (byte) ((float) pix[2] * origgrey / specgrey);
    }
}


/*
==================
GL_ProcessTexturePass

Runs a texture through the required pass for uploading it.
Up to 2 passes may be required - the first for the standard texture, which all textures
will go through.  The second is for lumas and only for textures where they are relevant.
==================
*/
static void GL_ProcessTexturePass (gltexture_t *glt, qboolean LumaPass)
{
    int				i, j;
    int				w, h;
    byte			*pix;
    byte			*data32 = NULL;
    unsigned int	*scaled = NULL;
    char			TexName[128];
    int				gltexnum;

    // save a pointer to the original data to preserve it as valid for the next pass
    byte			*olddata = glt->data;

    // lumas are only valid for mipmapped textures
    if (LumaPass && !(glt->flags & TEX_MIPMAP)) return;

    // this should never happen unless i fuck things up somewhere...
    if ((glt->flags & TEX_RESOURCE) && !(glt->flags & TEX_32BPP))
    {
        Sys_Error ("TEX_RESOURCE without TEX_32BPP");
    }

    // only bother with this if the texture doesn't come from a resource
    if (!(glt->flags & TEX_RESOURCE))
    {
        // look for an external version
        // (kill va madness and weirdness where multiple nested calls to va screw things up badly)
        if (LumaPass)
        {
            va_snprintf ("GL_ProcessTexturePass", TexName, sizeof(TexName), "%s_luma", glt->identifier);
        }
        else
        {
            va_snprintf ("GL_ProcessTexturePass", TexName, sizeof(TexName), "%s", glt->identifier);
        }
        data32 = GL_TryForExternalTexture (glt, TexName);

        // if we got one, set the data pointer.
        if (data32)
        {
            glt->data = data32;

            // set the luma flag if we're in the luma pass
            if (LumaPass)
            {
                glt->flags |= TEX_HASLUMA;
            }
        }

        // check for the luma flag in a luma pass.  this will have been either set in the standard pass for an 8bpp or
        // just up above for a 32bpp
        if (LumaPass && !(glt->flags & TEX_HASLUMA)) return;

        // check for already at 32bpp
        if (!(glt->flags & TEX_32BPP) && !data32)
        {
            // make sure we have texture data!!!
            if (!glt->data) return;

            // 8 bit texture so expand to 32 bpp
            // first, allocate a new data buffer
            data32 = malloc (glt->width * glt->height * 4);

            // convert into the new data buffer
            GL_Convert8BPPTo32BPP (glt, (unsigned int *) data32, LumaPass);

            // switch the data pointer
            glt->data = data32;
        }
    }

    // get power of 2 dimensions
    for (w = 1; w < glt->width; w *= 2);
    for (h = 1; h < glt->height; h *= 2);

    // JESUS!  I never tested this in the old one...
    if (w > GLCaps.GL_MaxTextureSize) w = GLCaps.GL_MaxTextureSize;
    if (h > GLCaps.GL_MaxTextureSize) h = GLCaps.GL_MaxTextureSize;

    // now check for resizing
    if (w != glt->width || h != glt->height)
    {
        // allocated a buffer to hold the resampled data
        scaled = (unsigned int *) malloc (w * h * sizeof (unsigned int));

        // resize the texture
        GL_ResampleTexture ((unsigned int *) glt->data, glt->width, glt->height, scaled, w, h);

        // switch the data pointer
        glt->data = (byte *) scaled;
    }

    // do our various hack jobs here...
    if (LumaPass) GL_QuantumLumaHackery (glt, glt->data);
    if (glt->flags & TEX_ALPHAMASK) GL_AlphaMaskTexture (glt, glt->data);
    if (glt->flags & TEX_LIQUID) GL_SetLiquidColour (glt, glt->data);

    // check for the presence of alpha pixels
    // this is probably a redundant and unnecessary test cos openGL stores in RGBA internally anyway...
    // as well as the alpha check, this derives a normal map scale to help keep the effect consistent across different texes
    pix = glt->data;
    brightest = -1;
    darkest = 256;

    for (i = 0; i < w; i++)
    {
        for (j = 0; j < h; j++, pix += 4)
        {
            int temp = ((int) pix[0] * 30 + (int) pix[1] * 59 + (int) pix[2] * 11) / 100;

            // find the brightest and the darkest
            if (temp > brightest)
            {
                brightest = temp;
            }

            if (temp < darkest)
            {
                darkest = temp;
            }

            // do the alpha check at 250 to catch any rogue errors
            // (that's rogue as in "rogue elephant", *not* rogue as in DOE, by the way...)
            if (pix[3] < 250)
            {
                glt->flags |= TEX_HASALPHA;
            }
        }
    }

    // subtract darkest from brightest to get the scale
    brightest -= darkest;

    // don't allow too small (only valid if a tex is all the same colour)
    if (brightest < 1)
    {
        brightest = 1;
    }

    // batch up calculations
    brightest = 1 / brightest;
    brightest *= 255;

    // make a normal map before uploading the texture so that the scaling doesn't affect it
    if (!LumaPass && (glt->flags & TEX_MAKENORMAL))
    {
        MakeDOT3 (glt, (unsigned int *) glt->data, w, h);

        // now upscale it to preserve the correct colour range
        pix = glt->data;

        for (i = 0; i < w; i++)
        {
            for (j = 0; j < h; j++, pix += 4)
            {
                int k;

                for (k = 0; k < 3; k++)
                {
                    float temp = pix[k];

                    temp *= BUMPRESCALE;

                    if (temp > 255)
                    {
                        temp = 255;
                    }
                    pix[k] = (byte) temp;
                }
            }
        }

        // fake specular highlights
        MakeSpec (glt, (unsigned int *) glt->data, w, h);
    }

    // upload now
    gltexnum = GL_UploadTextureToOpenGL (glt->data, w, h, glt->flags);

    // set the correct texture number, unless we're in TEX_NOGEN mode in which case we're not bothered with doing so
    if (LumaPass && !(glt->flags & TEX_NOGEN))
    {
        glt->lumatexnum = gltexnum;
    }
    else if (!(glt->flags & TEX_NOGEN))
    {
        glt->gltexnum = gltexnum;
    }

    // free the resized memory if it was allocated
    if (scaled)
    {
        free (scaled);
    }

    // free the 32bpp memory if it was allocated
    if (data32)
    {
        free (data32);
    }

    // restore the original data pointer
    glt->data = olddata;

    // mark it as loaded
    glt->flags |= TEX_LOADEDOK;
}


/*
================
GL_GetTextureSlotInCache

Returns a valid texture slot, either from the cache or a new one.  If it's from the cache, no further
processing is needed.  If it's a new one, it needs to go through the loading routines.
================
*/
gltexture_t *GL_GetTextureSlotInCache (char *identifier, byte *data, int width, int height)
{
    int			i;
    gltexture_t *glt;
    gltexture_t *freeslot = NULL;
    int			oldest = 0;
    int			oldestage = -1;
    byte		md5digest[16];

    // make an MD5 of the texture data.  we'll always need this for either cache comparisons or for
    // storing the MD5 of a new texture, so it's safe to do it here
    MD5_Checksum (data, width * height, md5digest);

    // replace invalid characters in the texture name prior to doing anything else
    for (i = 0; i < (int)strlen (identifier); i++)
    {
        if (identifier[i] == '*')
        {
            identifier[i] = '#';
        }

        if (identifier[i] == '.')
        {
            identifier[i] = 0;
            break;
        }
    }

    // check the cache
    for (i = 0; i < MAX_GL_TEXTURES; i++)
    {
        glt = &texturecache[i];

        // store the oldest.
        // the TEX_BSP condition prevents model and other textures from
        // being bumped, cos they're always needed.
        if (glt->textureage > oldestage && (glt->flags & TEX_BSP))
        {
            oldest = i;
            oldestage = glt->textureage;
        }

        // free slots have no texture number set
        if (!glt->gltexnum)
        {
            // save out the first free slot we hit
            if (!freeslot)
            {
                freeslot = glt;
            }

            // because this is a free slot we don't need to check the cache
            continue;
        }

        // check the cache
        // restrict it to the md5 cos identifiers can sometimes mismatch, and width and height can be changed
        // by the loading process.
        if (MD5_Compare (md5digest, glt->md5digest))
        {
            // checksums match.
            // set the age to 0 cos it's just been used
            glt->textureage = 0;

            // tell the loader that this texture came from the cache (meaning that it's data
            // pointer is likely to be invalid, and that certain additional operations are
            // likely to have already been performed, so there's no need to do them again).
            glt->flags |= TEX_INCACHE;

            // return the texture from the cache
            return glt;
        }
    }

    // oh dear.  we have no more free slots left, so bump the oldest texture.
    // if the oldest age is 0, we have serious trouble, but this is a highly unlikely occurrance.
    // we still have a potentially active "undefined behaviour" condition here though...
    if (!freeslot) freeslot = &texturecache[oldest];

    // if the free slot contains texture objects, free them.
    // this should only happen if the oldest is bumped.
    if (freeslot->gltexnum) glDeleteTextures (1, &freeslot->gltexnum);
    if (freeslot->lumatexnum) glDeleteTextures (1, &freeslot->lumatexnum);
    if (freeslot->normtexnum) glDeleteTextures (1, &freeslot->normtexnum);

    // init the texture slot with default values
    GL_InitTextureSlot (freeslot);

    // store the texture properties we have right now
    memcpy (freeslot->identifier, identifier, sizeof(freeslot->identifier));
    freeslot->width = width;
    freeslot->height = height;

    // store the md5
    MD5_Copy (md5digest, freeslot->md5digest);

    // set the data pointer for loading
    freeslot->data = data;

    // init the flags to nothing, we'll set these properly in the loader
    freeslot->flags = TEX_TEXTURE;

    // return the gltexture_t we just loaded
    return freeslot;
}


/*
==================
GL_LoadQuakeTexture

Texture loading interface.

Much cleaned up.  This might even be a viable single entry point (AT LAST!!!)

2 ways of calling:
	GL_LoadQuakeTexture (mytexture, etc...
	Load a texture straight up into a predefined gltexture_t struct

	mytexture = GL_LoadQuakeTexture (NULL, etc...
	Take a texture slot from the cache
==================
*/
gltexture_t *GL_LoadQuakeTexture (gltexture_t *glt, char *identifier, void *data, int width, int height, unsigned int flags)
{
    if (flags & TEX_LIGHTMAP)
    {
        // somebody shoot the programmer!
        if (!glt)
        {
            Sys_Error ("TEX_LIGHTMAP upload without a valid glt!");
        }

        // this lightmap is already allocated as a texture, so no need to reallocate
        if (glt->gltexnum)
        {
            return glt;
        }

        // upload to opengl (make sure it doesn't go in with any funny flags)
        glt->gltexnum = GL_UploadTextureToOpenGL (data, width, height, TEX_LIGHTMAP);

        // upload the luma as well for stains
        glt->lumatexnum = GL_UploadTextureToOpenGL (data, width, height, TEX_LIGHTMAP);

        // no need to do anything more here
        return glt;
    }

    // properly identify liquid textures
    if (identifier[0] == '*')
    {
        flags |= TEX_LIQUID;
    }

    if (!glt)
    {
        // we don't have a glt struct so we need to get one from the texture cache
        // if this texture already matches one in the cache, we just return, otherwise we go through the upload
        glt = GL_GetTextureSlotInCache (identifier, data, width, height);

        // took a texture from the cache
        if (glt->flags & TEX_INCACHE)
        {
            return glt;
        }
    }
    else
    {
        // predefined glt struct, so fill it in
        // clear to nothing
        GL_InitTextureSlot (glt);

        // set the base data
        glt->width = width;
        glt->height = height;

        // the "data" pointer is temporary and may be undefined following a map reload
        glt->data = data;

        // copy the identifier
        memcpy (glt->identifier, identifier, sizeof(glt->identifier));
    }

    // give it an md5 cos the init above destroys it
    MD5_Checksum (data, width * height, glt->md5digest);

    // copy the correct flags irrespective of whether it came from the cache or not
    glt->flags = flags;

    // first pass for the standard texture
    // this will also handle the generation of normal maps where required
    GL_ProcessTexturePass (glt, false);

    // failed to load!!!  yikes!!!
    // (use this in the calling func for loading a texture where we have no internal representation and we must
    // have an external one only - i.e. mapshots, skyboxes, etc.)
    if (!(glt->flags & TEX_LOADEDOK))
    {
        return NULL;
    }

    // second pass for the luma component
    GL_ProcessTexturePass (glt, true);

    return glt;
}


/*
================
GL_LoadBitmapResourceTexture

Loads a BMP resource texture

only bitmaps can be loaded from resources, cos versions of Windows prior to XP don't support other
image formats in resources.

This is the only entry point to texture loading aside from GL_LoadQuakeTexture, and is only used because
resource textures are done very differently.
================
*/
qboolean GL_LoadBitmapResourceTexture (gltexture_t *glt, int ResourceID, unsigned int flags)
{
    unsigned char	*ImageRGBAData;
    int				i, j;
    int				run;
    int				runlength;
    unsigned char	*bmIn;
    BITMAPINFO		*bmInfo;
    RGBQUAD			bmPalette[256];
    RGBQUAD			*PaletteIn;
    unsigned char	*bmData;
    unsigned char	*bmRaw;
    int				pad;
    HINSTANCE		hInstance;
    HANDLE			hBitmap;

    // get a handle to the Realm executable
    hInstance = GetModuleHandle (NULL);

    // load the resource
    hBitmap = LoadResource (hInstance, FindResource (hInstance, MAKEINTRESOURCE (ResourceID), RT_BITMAP));

    // lock it
    bmRaw = (byte *) LockResource (hBitmap);

    // get the bm info header
    // read the header
    bmInfo = (BITMAPINFO *) bmRaw;

    // skip past the header (for 24 BPP RGB images the pixel data immediately follows, otherwise a palette is next)
    bmRaw += sizeof (BITMAPINFO);

    if (bmInfo->bmiHeader.biCompression != BI_RGB)
    {
        if (!(bmInfo->bmiHeader.biCompression == BI_RLE8 && bmInfo->bmiHeader.biBitCount == 8))
        {
            goto BadBMP;
        }
    }

    // support 4, 8 and 24 BPP modes
    if (bmInfo->bmiHeader.biBitCount != 4 &&
            bmInfo->bmiHeader.biBitCount != 8 &&
            bmInfo->bmiHeader.biBitCount != 24) goto BadBMP;

    ImageRGBAData = malloc (bmInfo->bmiHeader.biWidth * abs (bmInfo->bmiHeader.biHeight) * 4);
    bmData = ImageRGBAData;

    if (bmInfo->bmiHeader.biBitCount != 24)
    {
        // read the palette entries
        // the first one is in the BITMAPINFO struct
        bmPalette[0].rgbBlue = bmInfo->bmiColors[0].rgbBlue;
        bmPalette[0].rgbGreen = bmInfo->bmiColors[0].rgbGreen;
        bmPalette[0].rgbRed = bmInfo->bmiColors[0].rgbRed;
        bmPalette[0].rgbReserved = bmInfo->bmiColors[0].rgbReserved;

        // remaining entries follow the struct
        for (i = 1; i < (int)bmInfo->bmiHeader.biClrUsed; i++)
        {
            PaletteIn = (RGBQUAD *) bmRaw;

            bmPalette[i].rgbBlue = PaletteIn->rgbBlue;
            bmPalette[i].rgbGreen = PaletteIn->rgbGreen;
            bmPalette[i].rgbRed = PaletteIn->rgbRed;
            bmPalette[i].rgbReserved = PaletteIn->rgbReserved;

            bmRaw += sizeof (RGBQUAD);
        }
    }

    // each row in a bitmap is padded to align it on a 4 unsigned char boundary, so find
    // out how much padding we need (this may not work with RLE images)
    for (i = 0; ; i++)
    {
        pad = i * 4;
        if (pad >= (bmInfo->bmiHeader.biWidth * 3)) break;
    }

    pad -= bmInfo->bmiHeader.biWidth * 3;

    // load the data
    for (i = 0; i < bmInfo->bmiHeader.biWidth * abs (bmInfo->bmiHeader.biHeight); i++)
    {
        if (bmInfo->bmiHeader.biBitCount == 4)
        {
            bmIn = bmRaw++;

            bmData[0] = bmPalette[*bmIn >> 4].rgbRed;
            bmData[1] = bmPalette[*bmIn >> 4].rgbGreen;
            bmData[2] = bmPalette[*bmIn >> 4].rgbBlue;
            bmData[3] = 255;

            bmData[4] = bmPalette[*bmIn % 16].rgbRed;
            bmData[5] = bmPalette[*bmIn % 16].rgbGreen;
            bmData[6] = bmPalette[*bmIn % 16].rgbBlue;
            bmData[7] = 255;

            bmData += 8;

            // increment i again cos we're reading two entires at a time
            i++;

            continue;
        }
        else if (bmInfo->bmiHeader.biBitCount == 8)
        {
            if (bmInfo->bmiHeader.biCompression == BI_RLE8)
            {
                bmIn = bmRaw++;

                if (*bmIn == 0)
                {
                    bmIn = bmRaw++;
                    i++;

                    switch (*bmIn)
                    {
                    case 0:
                        // End of Line - just do nothing - trust me, it works.
                        break;
                    case 1:
                        // End of Bitmap
                        goto bmDone;
                        break;
                    case 2:
                        // Delta
                        // (tests indicate this isn't actually used... most likely cos the bitmap format
                        // doesn't specify what the skipped pixels are to be filled with!!!)
                        bmIn = bmRaw++;
                        i++;

                        bmData += *bmIn * 4;

                        bmIn = bmRaw++;
                        i++;

                        bmData += *bmIn * bmInfo->bmiHeader.biWidth * 4;

                        break;
                    default:
                        // absolute mode
                        runlength = *bmIn;

                        for (run = 0; run < runlength; run++)
                        {
                            bmIn = bmRaw++;
                            i++;

                            bmData[0] = bmPalette[*bmIn].rgbRed;
                            bmData[1] = bmPalette[*bmIn].rgbGreen;
                            bmData[2] = bmPalette[*bmIn].rgbBlue;
                            bmData[3] = 255;

                            bmData += 4;
                        }

                        if (runlength % 2)
                        {
                            // align on word boundaries
                            bmIn = bmRaw++;
                            i++;
                        }

                        break;
                    }
                }
                else
                {
                    // encoded mode
                    runlength = *bmIn;

                    bmIn = bmRaw++;
                    i++;

                    for (run = 0; run < runlength; run++)
                    {
                        bmData[0] = bmPalette[*bmIn].rgbRed;
                        bmData[1] = bmPalette[*bmIn].rgbGreen;
                        bmData[2] = bmPalette[*bmIn].rgbBlue;
                        bmData[3] = 255;

                        bmData += 4;
                    }
                }
                continue;
            }
            bmIn = bmRaw++;

            bmData[0] = bmPalette[*bmIn].rgbRed;
            bmData[1] = bmPalette[*bmIn].rgbGreen;
            bmData[2] = bmPalette[*bmIn].rgbBlue;
            bmData[3] = 255;

            bmData += 4;

            continue;
        }
        else
        {
            // 24 BPP
            bmData[2] = *bmRaw++;
            bmData[1] = *bmRaw++;
            bmData[0] = *bmRaw++;
            bmData[3] = 255;

            bmData += 4;
        }

        // read the padding
        for (j = 0; j < pad; j++)
        {
            bmRaw++;
        }
    }

bmDone:
    ;

    // check orientation
    if (bmInfo->bmiHeader.biHeight > 0)
    {
        // bottom-up bitmap so we must invert it (sigh)
        for (i = 0; i < abs (bmInfo->bmiHeader.biHeight) / 2; i++)
        {
            for (j = 0; j < bmInfo->bmiHeader.biWidth; j++)
            {
                unsigned int srcpos = (i * bmInfo->bmiHeader.biWidth) + j;
                unsigned int dstpos = (((abs (bmInfo->bmiHeader.biHeight) - 1) - i) * bmInfo->bmiHeader.biWidth) + j;
                unsigned int tmp;

                tmp = ((unsigned int *) ImageRGBAData)[dstpos];
                ((unsigned int *) ImageRGBAData)[dstpos] = ((unsigned int *) ImageRGBAData)[srcpos];
                ((unsigned int *) ImageRGBAData)[srcpos] = tmp;
            }
        }
    }

    // fill in properties
    GL_InitTextureSlot (glt);

    glt->width = bmInfo->bmiHeader.biWidth;
    glt->height = abs (bmInfo->bmiHeader.biHeight);

    // copy the flags - make sure it's properly flagged as 32 bit
    glt->flags = (flags | TEX_32BPP | TEX_RESOURCE);

    // set the data pointer
    glt->data = ImageRGBAData;

    // one pass cos there should never have lumas
    GL_ProcessTexturePass (glt, false);
    free (ImageRGBAData);
    return true;
BadBMP:
    ;
    return false;
}


/*
============================================================================================================================================

			TEXTURE CACHE MANAGEMENT

============================================================================================================================================
*/

/*
=================
GL_InitTextureSlot

Initializes a texture slot with the default values.  Called at startup, and also whenever a texture
slot is created or replaced in the cache, and also when a texture is nulled through ageing.
=================
*/
void GL_InitTextureSlot (gltexture_t *glt)
{
    // clear to 0
    memset (glt, 0, sizeof (gltexture_t));

    // and again, to be sure, to be sure
    glt->data = NULL;
    glt->gltexnum = 0;
    glt->height = 0;
    glt->identifier[0] = 0;
    glt->liquidtype = LIQUIDTYPE_NOTLIQUID;
    glt->lumatexnum = 0;
    glt->normtexnum = 0;
    glt->md5digest[0] = 0;
    glt->rgba[0] = glt->rgba[1] = glt->rgba[2] = glt->rgba[3] = 0.0f;
    glt->textureage = 0;
    glt->width = 0;
}


void GL_InitTextureCache (void)
{
    int i;

    for (i = 0; i < MAX_GL_TEXTURES; i++)
    {
        GL_InitTextureSlot (&texturecache[i]);
    }
}


/*
================
GL_TextureAging

Keeps and updates a record of the textures ages, so that the oldest texture can be correctly
bumped if the texture cache gets full.
================
*/
void GL_TextureAging (void)
{
    int i;
    gltexture_t	*glt;
    float TexPriorities[2] = {1, 0};

    for (i = 0; i < MAX_GL_TEXTURES; i++)
    {
        // get the current entry
        glt = &texturecache[i];

        // not a texture yet
        if (!glt->gltexnum) continue;

        // set texture priorities
        if (glt->textureage < 3)
        {
            glPrioritizeTextures (1, &glt->gltexnum, &TexPriorities[0]);
            if (glt->lumatexnum) glPrioritizeTextures (1, &glt->lumatexnum, &TexPriorities[0]);
            if (glt->normtexnum) glPrioritizeTextures (1, &glt->normtexnum, &TexPriorities[0]);
        }
        else if (glt->textureage < 10)
        {
            glPrioritizeTextures (1, &glt->gltexnum, &TexPriorities[1]);
            if (glt->lumatexnum) glPrioritizeTextures (1, &glt->lumatexnum, &TexPriorities[1]);
            if (glt->normtexnum) glPrioritizeTextures (1, &glt->normtexnum, &TexPriorities[1]);
        }

        // explicit no aging
        if (glt->flags & TEX_NOAGE) continue;

        // bsp only
        if (!(glt->flags & TEX_BSP)) continue;

        // age the texture
        glt->textureage++;

        // if it's too old, destroy it
        if (glt->textureage > 11)
        {
            glDeleteTextures (1, &glt->gltexnum);

            if (glt->lumatexnum) glDeleteTextures (1, &glt->lumatexnum);
            if (glt->normtexnum) glDeleteTextures (1, &glt->normtexnum);

            GL_InitTextureSlot (glt);
        }
    }
}


/*
================
GL_FlushTextureCache

Flushes all map textures from the cache (non map textures are left alone)
================
*/
void GL_FlushTextureCache (void)
{
    int i;
    gltexture_t	*glt;

    for (i = 0; i < MAX_GL_TEXTURES; i++)
    {
        // get the current entry
        glt = &texturecache[i];

        // always keep anything that ain't a map texture
        if (!(glt->flags & TEX_BSP)) continue;

        // explicitly don't flush it
        if (glt->flags & TEX_NOAGE) continue;

        // not a texture yet
        if (!glt->gltexnum) continue;

        glDeleteTextures (1, &glt->gltexnum);

        if (glt->lumatexnum) glDeleteTextures (1, &glt->lumatexnum);
        if (glt->normtexnum) glDeleteTextures (1, &glt->normtexnum);

        GL_InitTextureSlot (glt);
    }
}

