/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// in_win.c -- windows 95 mouse and joystick code
// 02/21/97 JCB Added extended DirectInput code to support external controllers.

#include "quakedef.h"
#include "winquake.h"

// mouse variables
cvar_t	m_filter = {"m_filter","0"};

int			mouse_buttons;
int			mouse_oldbuttonstate;
POINT		current_pos;
int			mouse_x, mouse_y, old_mouse_x, old_mouse_y, mx_accum, my_accum;

static qboolean	restore_spi;
static int		originalmouseparms[3], newmouseparms[3] = {0, 0, 1};

unsigned int	uiWheelMessage;
qboolean		mouseactive;
qboolean		mouseinitialized;
static qboolean	mouseparmsvalid, mouseactivatetoggle;
static qboolean	mouseshowtoggle = 1;

cvar_t in_mmlook = {"in_mlook", "1", true};

/*
===========
Force_CenterView_f
===========
*/
void Force_CenterView_f (void)
{
    cl.viewangles[PITCH] = 0;
}

/*
===========
IN_UpdateClipCursor
===========
*/
void IN_UpdateClipCursor (void)
{
    if (mouseinitialized && mouseactive)
    {
        ClipCursor (&window_rect);
    }
}

/*
===========
IN_ShowMouse
===========
*/
void IN_ShowMouse (void)
{
    if (!mouseshowtoggle)
    {
        ShowCursor (TRUE);
        mouseshowtoggle = 1;
    }
}


/*
===========
IN_HideMouse
===========
*/
void IN_HideMouse (void)
{
    if (mouseshowtoggle)
    {
        ShowCursor (FALSE);
        mouseshowtoggle = 0;
    }
}


/*
===========
IN_ActivateMouse
===========
*/
void IN_ActivateMouse (void)
{
    mouseactivatetoggle = true;

    if (mouseinitialized)
    {
        if (mouseparmsvalid)
        {
            restore_spi = SystemParametersInfo (SPI_SETMOUSE, 0, newmouseparms, 0);
        }
        SetCursorPos (window_center_x, window_center_y);
        SetCapture (mainwindow);
        ClipCursor (&window_rect);

        mouseactive = true;
    }
}


/*
===========
IN_SetQuakeMouseState
===========
*/
void IN_SetQuakeMouseState (void)
{
    if (mouseactivatetoggle)
    {
        IN_ActivateMouse ();
    }
}


/*
===========
IN_DeactivateMouse
===========
*/
void IN_DeactivateMouse (void)
{
    mouseactivatetoggle = false;

    if (mouseinitialized)
    {
        if (restore_spi)
        {
            SystemParametersInfo (SPI_SETMOUSE, 0, originalmouseparms, 0);
        }
        ClipCursor (NULL);
        ReleaseCapture ();

        mouseactive = false;
    }
}


/*
===========
IN_RestoreOriginalMouseState
===========
*/
void IN_RestoreOriginalMouseState (void)
{
    if (mouseactivatetoggle)
    {
        IN_DeactivateMouse ();
        mouseactivatetoggle = true;
    }

    // try to redraw the cursor so it gets reinitialized, because sometimes it
    // has garbage after the mode switch
    ShowCursor (TRUE);
    ShowCursor (FALSE);
}


/*
===========
IN_StartupMouse
===========
*/
void IN_StartupMouse (void)
{
    if ( COM_CheckParm ("-nomouse") ) return;

    mouseinitialized = true;

    mouseparmsvalid = SystemParametersInfo (SPI_GETMOUSE, 0, originalmouseparms, 0);

    if (mouseparmsvalid)
    {
        if ( COM_CheckParm ("-noforcemspd") )
            newmouseparms[2] = originalmouseparms[2];

        if ( COM_CheckParm ("-noforcemaccel") )
        {
            newmouseparms[0] = originalmouseparms[0];
            newmouseparms[1] = originalmouseparms[1];
        }

        if ( COM_CheckParm ("-noforcemparms") )
        {
            newmouseparms[0] = originalmouseparms[0];
            newmouseparms[1] = originalmouseparms[1];
            newmouseparms[2] = originalmouseparms[2];
        }
    }
    mouse_buttons = 3;

    // if a fullscreen video mode was set before the mouse was initialized,
    // set the mouse state appropriately
    if (mouseactivatetoggle)
    {
        IN_ActivateMouse ();
    }
}


/*
===========
IN_Init
===========
*/
void IN_Init (void)
{
    // mouse variables
    Cvar_RegisterVariable (&m_filter);
    Cvar_RegisterVariable (&in_mmlook);

    Cmd_AddCommand ("force_centerview", Force_CenterView_f);

    uiWheelMessage = RegisterWindowMessage ( "MSWHEEL_ROLLMSG" );

    IN_StartupMouse ();
}

/*
===========
IN_Shutdown
===========
*/
void IN_Shutdown (void)
{

    IN_DeactivateMouse ();
    IN_ShowMouse ();
}


/*
===========
IN_MouseEvent
===========
*/
void IN_MouseEvent (int mstate)
{
    int	i;

    if (mouseactive)
    {
        // perform button actions
        for (i=0 ; i<mouse_buttons ; i++)
        {
            if ( (mstate & (1<<i)) &&
                    !(mouse_oldbuttonstate & (1<<i)) )
            {
                Key_Event (K_MOUSE1 + i, true);
            }

            if ( !(mstate & (1<<i)) &&
                    (mouse_oldbuttonstate & (1<<i)) )
            {
                Key_Event (K_MOUSE1 + i, false);
            }
        }
        mouse_oldbuttonstate = mstate;
    }
}


/*
===========
IN_MouseMove
===========
*/
void IN_MouseMove (usercmd_t *cmd)
{
    int					mx, my;

    if (!mouseactive) return;

    GetCursorPos (&current_pos);
    mx = current_pos.x - window_center_x + mx_accum;
    my = current_pos.y - window_center_y + my_accum;
    mx_accum = 0;
    my_accum = 0;

    if (m_filter.value)
    {
        mouse_x = (mx + old_mouse_x) * 0.5;
        mouse_y = (my + old_mouse_y) * 0.5;
    }
    else
    {
        mouse_x = mx;
        mouse_y = my;
    }

    old_mouse_x = mx;
    old_mouse_y = my;

    mouse_x *= sensitivity.value;
    mouse_y *= sensitivity.value;

    // add mouse X/Y movement to cmd
    if ( (in_strafe.state & 1) || (lookstrafe.value && (in_mmlook.value) ))
    {
        cmd->sidemove += m_side.value * mouse_x;
    }
    else
    {
        cl.viewangles[YAW] -= m_yaw.value * mouse_x;
    }

    if (in_mmlook.value)
    {
        V_StopPitchDrift ();
    }

    if ( (in_mmlook.value) && !(in_strafe.state & 1))
    {
        cl.viewangles[PITCH] += m_pitch.value * mouse_y;

        if (cl.viewangles[PITCH] > 80)
        {
            cl.viewangles[PITCH] = 80;
        }

        if (cl.viewangles[PITCH] < -70)
        {
            cl.viewangles[PITCH] = -70;
        }
    }
    else
    {
        if ((in_strafe.state & 1) && noclip_anglehack)
        {
            cmd->upmove -= m_forward.value * mouse_y;
        }
        else
        {
            cmd->forwardmove -= m_forward.value * mouse_y;
        }
    }

    // if the mouse has moved, force it to the center, so there's room to move
    if (mx || my)
    {
        SetCursorPos (window_center_x, window_center_y);
    }
}


/*
===========
IN_Move
===========
*/
void IN_Move (usercmd_t *cmd)
{
    if (ActiveApp && !Minimized)
    {
        IN_MouseMove (cmd);
    }
}


/*
===========
IN_Accumulate
===========
*/
void IN_Accumulate (void)
{
    if (mouseactive)
    {
        GetCursorPos (&current_pos);

        mx_accum += current_pos.x - window_center_x;
        my_accum += current_pos.y - window_center_y;

        // force the mouse to the center, so there's room to move
        SetCursorPos (window_center_x, window_center_y);
    }
}


/*
===================
IN_ClearStates
===================
*/
void IN_ClearStates (void)
{
    if (mouseactive)
    {
        mx_accum = 0;
        my_accum = 0;
        mouse_oldbuttonstate = 0;
    }
}
